from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "tier_1_support"
CALL_CONTROL_ID = "AgDIxmoRX6QMuaIj_uXRXnPAXP0QlNfXczRrZvZakpWxBlpw48KyZQ=="


class TestQueue(object):
    def test_is_retrievable(self, request_mock):
        telnyx.Queue.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/queues/%s" % TEST_RESOURCE_ID)

    @pytest.mark.skip(reason="Prism update")
    def test_can_call_retrieve_call(self, request_mock):
        resource = telnyx.Queue.retrieve(TEST_RESOURCE_ID)
        resource.retrieve_call(CALL_CONTROL_ID)
        request_mock.assert_requested(
            "get", "/v2/queues/%s/calls/%s" % (resource.name, CALL_CONTROL_ID)
        )

    @pytest.mark.skip(reason="Need live confirm")
    def test_can_call_list_calls(self, request_mock):
        resource = telnyx.Queue.retrieve(TEST_RESOURCE_ID)
        queue_calls = resource.list_calls()
        request_mock.assert_requested("get", "/v2/queues/%s/calls" % resource.name)
        assert isinstance(queue_calls.data[0], telnyx.QueueCall)



from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "sms", path="/v2/verifications/sms", operations=["create"]
)
@nested_resource_class_methods(
    "psd2", path="/v2/verifications/psd2", operations=["create"]
)
@nested_resource_class_methods(
    "call", path="/v2/verifications/call", operations=["create"]
)
@nested_resource_class_methods(
    "flashcall", path="/v2/verifications/flashcall", operations=["create"]
)
@nested_resource_class_methods(
    "whatsapp", path="/v2/verifications/whatsapp", operations=["create"]
)
@nested_resource_class_methods(
    "verify_by_phone_number", path="by_phone_number/{phone_number}/actions/verify", operations=["create"]
)
@nested_resource_class_methods(
    "verify_by_id", path="actions/verify", operations=["create"]
)
@nested_resource_class_methods(
    "by_phone_number", path="by_phone_number/{phone_number}", operations=["retrieve"]
)
class Verification(CreateableAPIResource, ListableAPIResource, UpdateableAPIResource):
    OBJECT_NAME = "verification"

    @classmethod
    def sms(cls, **params):
        return Verification.create_sms(None, **params)

    @classmethod
    def psd2(cls, **params):
        return Verification.create_psd2(None, **params)

    @classmethod
    def call(cls, **params):
        return Verification.create_call(None, **params)

    @classmethod
    def flashcall(cls, **params):
        return Verification.create_flashcall(None, **params)

    @classmethod
    def whatsapp(cls, **params):
        return Verification.create_whatsapp(None, **params)

    @classmethod
    def verify_by_phone_number(cls, phone_number, **params):
        return Verification.create_verify_by_phone_number(phone_number, **params)

    @classmethod
    def verify_by_id(cls, verification_id, **params):
        return Verification.create_verify_by_id(verification_id, **params)

    @classmethod
    def by_phone_number(cls, phone_number):
        return Verification.retrieve_by_phone_number(phone_number)



from __future__ import absolute_import, division, print_function

import telnyx

TEST_RESOURCE_ID = "12ade33a-21c0-473b-b055-b3c836e1c292"


class TestNumberOrder(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.NumberOrder.list()
        request_mock.assert_requested("get", "/v2/number_orders")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.NumberOrder)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.NumberOrder.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/number_orders/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.NumberOrder)

    def test_is_creatable(self, request_mock):
        resource = telnyx.NumberOrder.create(
            phone_numbers=[
                {"phone_number": "+12223334444", "regulatory_requirements": []}
            ],
            customer_reference="MY REF 001",
            connection_id="442191469269222625",
            messaging_profile_id="730911e3-8488-40a8-a818-dc0a5df8bc03",
        )
        request_mock.assert_requested("post", "/v2/number_orders")
        assert isinstance(resource, telnyx.NumberOrder)

    def test_is_saveable(self, request_mock):
        number_order = telnyx.NumberOrder.retrieve(TEST_RESOURCE_ID)
        number_order.customer_reference = "value"
        resource = number_order.save()
        request_mock.assert_requested(
            "patch", "/v2/number_orders/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.NumberOrder)
        assert resource is number_order

    def test_is_modifiable(self, request_mock):
        resource = telnyx.NumberOrder.modify(
            TEST_RESOURCE_ID, customer_reference="Test"
        )
        request_mock.assert_requested(
            "patch", "/v2/number_orders/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.NumberOrder)



from __future__ import absolute_import, division, print_function

import telnyx


class TestNotificationEventCondition(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.NotificationEventCondition.list()
        request_mock.assert_requested("get", "/v2/notification_event_conditions")
        assert isinstance(resources.data, list)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import CreateableAPIResource


class PresignedObjectUrl(CreateableAPIResource):
    OBJECT_NAME = "presigned_object_url"



from __future__ import absolute_import, division, print_function

import pytest

import telnyx


class TestNestedResourceClassMethods(object):
    @telnyx.api_resources.abstract.nested_resource_class_methods(
        "nested", operations=["create", "retrieve", "update", "delete", "list"]
    )
    class MainResource(telnyx.api_resources.abstract.APIResource):
        OBJECT_NAME = "mainresource"

    def test_nested_resource_class_methods_no_operator(self):
        with pytest.raises(ValueError):
            telnyx.api_resources.abstract.nested_resource_class_methods(
                "a", operations=None
            )

    def test_create_nested(self, request_mock):
        request_mock.stub_request(
            "post",
            "/v2/mainresources/id/nesteds",
            {"id": "nested_id", "object": "nested", "foo": "bar"},
        )
        nested_resource = self.MainResource.create_nested("id", foo="bar")
        request_mock.assert_requested(
            "post", "/v2/mainresources/id/nesteds", {"foo": "bar"}, None
        )
        assert nested_resource.foo == "bar"

    def test_retrieve_nested(self, request_mock):
        request_mock.stub_request(
            "get",
            "/v2/mainresources/id/nesteds/nested_id",
            {"id": "nested_id", "object": "nested", "foo": "bar"},
        )
        nested_resource = self.MainResource.retrieve_nested("id", "nested_id")
        request_mock.assert_requested(
            "get", "/v2/mainresources/id/nesteds/nested_id", {}, None
        )
        assert nested_resource.foo == "bar"

    def test_modify_nested(self, request_mock):
        request_mock.stub_request(
            "post",
            "/v2/mainresources/id/nesteds/nested_id",
            {"id": "nested_id", "object": "nested", "foo": "baz"},
        )
        nested_resource = self.MainResource.modify_nested("id", "nested_id", foo="baz")
        request_mock.assert_requested(
            "post", "/v2/mainresources/id/nesteds/nested_id", {"foo": "baz"}, None
        )
        assert nested_resource.foo == "baz"

    def test_delete_nested(self, request_mock):
        request_mock.stub_request(
            "delete",
            "/v2/mainresources/id/nesteds/nested_id",
            {"id": "nested_id", "object": "nested", "deleted": True},
        )
        nested_resource = self.MainResource.delete_nested("id", "nested_id")
        request_mock.assert_requested(
            "delete", "/v2/mainresources/id/nesteds/nested_id", {}, None
        )
        assert nested_resource.deleted is True

    def test_list_nesteds(self, request_mock):
        request_mock.stub_request(
            "get", "/v2/mainresources/id/nesteds", {"object": "list", "data": []}
        )
        nested_resource = self.MainResource.list_nesteds("id")
        request_mock.assert_requested("get", "/v2/mainresources/id/nesteds", {}, None)
        assert isinstance(nested_resource.data, list)

    def test_invalid_operation(self):
        class Foo:
            pass

        decorator = telnyx.api_resources.abstract.nested_resource_class_methods(
            "nested", operations=["foo"]
        )

        with pytest.raises(ValueError):
            decorator(Foo)

    def test_rewrite_reserved_words(self, request_mock):
        request_mock.stub_request(
            "post",
            "/v2/mainresources/id/nesteds",
            {"id": "nested_id", "object": "nested", "from": "foo"},
        )
        nested_resource = self.MainResource.create_nested("id", from_="foo")
        request_mock.assert_requested(
            "post", "/v2/mainresources/id/nesteds", {"from": "foo"}, None
        )
        assert nested_resource.from_ == "foo"



from __future__ import absolute_import, division, print_function

import json


class TelnyxResponse(object):
    def __init__(self, body, code, headers):
        self.body = body
        self.code = code
        self.headers = headers
        self.data = json.loads(body)

    @property
    def request_id(self):
        try:
            return self.headers["request-id"]
        except KeyError:
            return None



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "provision", path="actions/provision", operations=["create"]
)
class VirtualCrossConnect(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "virtual_cross_connect"

    def provision(self, **params):
        return self.create_provision(**params)



from __future__ import absolute_import, division, print_function

import pytest
import telnyx

TEST_RESOURCE_ID = "0ccc7b54-4df3-4bca-a65a-3da1ecc777f1"


class TestDynamicEmergencyAddress(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.DynamicEmergencyAddress.list()
        request_mock.assert_requested("get", "/v2/dynamic_emergency_addresses")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.DynamicEmergencyAddress)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_retrievable(self, request_mock):
        resource = telnyx.DynamicEmergencyAddress.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/dynamic_emergency_addresses/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.DynamicEmergencyAddress)

    def test_is_creatable(self, request_mock):
        telnyx.DynamicEmergencyAddress.create(
            administrative_area="IL",
            house_number="311",
            locality="Chicago",
            postal_code="60654",
            street_name="Superior",
            country_coude="US",
        )

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_deletable(self, request_mock):
        resource = telnyx.DynamicEmergencyAddress.retrieve(TEST_RESOURCE_ID)
        resource.delete()



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import UpdateableAPIResource


class InboundChannel(UpdateableAPIResource):
    OBJECT_NAME = "inbound_channels"

    def instance_url(self):
        return "/v2/phone_numbers/inbound_channels"

    @classmethod
    def modify(cls, **params):
        url = "/v2/phone_numbers/inbound_channels"
        return cls._modify(url, **params)

    @classmethod
    def retrieve(cls, api_key=None, **params):
        return super(InboundChannel, cls).retrieve(None, api_key, **params)

    def save_method(self):
        return "patch"



from __future__ import absolute_import, division, print_function

import json

import pytest

import telnyx


class TestListObject(object):
    @pytest.fixture
    def list_object(self):
        return telnyx.ListObject.construct_from(
            {
                "data": ["foo"],
                "meta": {"page": {"size": 1, "number": 1}},
                "url": "/my/path",
            },
            "mykey",
        )

    def assert_list_response(self, res):
        assert isinstance(res[0], telnyx.MessagingProfile)
        assert res[0].foo == "bar"

    def assert_dict_response(self, res):
        assert isinstance(res, telnyx.MessagingProfile)
        assert res.foo == "bar"

    def test_for_loop(self, list_object):
        seen = []

        for item in list_object:
            seen.append(item)

        assert seen == ["foo"]

    def test_list(self, request_mock, list_object):
        request_mock.stub_request(
            "get",
            "/my/path",
            {"data": [{"record_type": "messaging_profile", "foo": "bar"}]},
        )

        res = list_object.list(myparam="you")

        request_mock.assert_requested("get", "/my/path", {"myparam": "you"}, None)
        self.assert_list_response(res.data)

    def test_create(self, request_mock, list_object):
        request_mock.stub_request(
            "post",
            "/my/path",
            {"data": {"record_type": "messaging_profile", "foo": "bar"}},
        )

        res = list_object.create(myparam="eter")

        request_mock.assert_requested("post", "/my/path", {"myparam": "eter"}, None)
        self.assert_dict_response(res)

    def test_retrieve(self, request_mock, list_object):
        request_mock.stub_request(
            "get",
            "/my/path/myid",
            {"data": {"record_type": "messaging_profile", "foo": "bar"}},
        )

        res = list_object.retrieve("myid", myparam="cow")

        request_mock.assert_requested("get", "/my/path/myid", {"myparam": "cow"}, None)

        self.assert_dict_response(res)

    def test_len(self, list_object):
        assert len(list_object) == 1

    def test_bool(self, list_object):
        assert list_object

        empty = telnyx.ListObject.construct_from({"data": []}, "mykey")
        assert bool(empty) is False

    def test_serialize_empty_list(self):
        empty = telnyx.ListObject.construct_from({"data": []}, "mykey")
        serialized = str(empty)
        deserialized = telnyx.ListObject.construct_from(json.loads(serialized), "mykey")
        assert deserialized == empty

    def test_serialize_nested_empty_list(self):
        empty = telnyx.ListObject.construct_from({"data": []}, "mykey")
        obj = telnyx.telnyx_object.TelnyxObject.construct_from(
            {"nested": empty}, "mykey"
        )
        serialized = str(obj)
        deserialized = telnyx.telnyx_object.TelnyxObject.construct_from(
            json.loads(serialized), "mykey"
        )
        assert deserialized.nested == empty


class TestAutoPaging:
    @staticmethod
    def pageable_model_response(ids, has_more, has_less, url):
        if has_more:
            total_pages = 2
            page_number = 1
        elif has_less:
            total_pages = 2
            page_number = 2
        else:
            total_pages = 1
            page_number = 1

        return {
            "data": [{"id": id, "record_type": "pageablemodel"} for id in ids],
            "meta": {"page_number": page_number, "total_pages": total_pages},
            "url": url,
        }

    @staticmethod
    def pageable_token_model_response(ids, token, url):
        return {
            "data": [{"id": id, "record_type": "pageablemodel"} for id in ids],
            "meta": {"next_page_token": token},
            "url": url,
        }

    def test_iter_one_page(self, request_mock):
        url = "/v2/pageablemodels"
        lo = telnyx.ListObject.construct_from(
            self.pageable_model_response(["pm_123", "pm_124"], False, False, url),
            "mykey",
        )

        request_mock.assert_no_request()

        seen = [item["id"] for item in lo.auto_paging_iter()]

        assert seen == ["pm_123", "pm_124"]

    def test_iter_one_page_by_token(self, request_mock):
        url = "/v2/pageablemodels"
        lo = telnyx.ListObject.construct_from(
            self.pageable_token_model_response(["pm_123", "pm_124"], None, url), "mykey"
        )

        request_mock.assert_no_request()

        seen = [item["id"] for item in lo.auto_paging_iter_by_token()]

        assert seen == ["pm_123", "pm_124"]

    def test_iter_two_pages(self, request_mock):
        url = "/v2/pageablemodels"
        lo = telnyx.ListObject.construct_from(
            self.pageable_model_response(["pm_123", "pm_124"], True, False, url),
            "mykey",
        )
        lo._retrieve_params = {"foo": "bar"}

        request_mock.stub_request(
            "get",
            "/v2/pageablemodels",
            self.pageable_model_response(["pm_125", "pm_126"], False, False, url),
        )

        seen = [item["id"] for item in lo.auto_paging_iter()]

        request_mock.assert_requested(
            "get",
            "/v2/pageablemodels",
            {"page": {"size": 20, "number": 2}, "foo": "bar"},
            None,
        )

        assert seen == ["pm_123", "pm_124", "pm_125", "pm_126"]

    def test_iter_by_token_two_pages(self, request_mock):
        url = "/v2/pageablemodels"
        lo = telnyx.ListObject.construct_from(
            self.pageable_token_model_response(["pm_123", "pm_124"], "123", url),
            "mykey",
        )
        lo._retrieve_params = {"foo": "bar"}

        request_mock.stub_request(
            "get",
            "/v2/pageablemodels",
            self.pageable_token_model_response(["pm_125", "pm_126"], None, url),
        )

        seen = [item["id"] for item in lo.auto_paging_iter_by_token()]

        request_mock.assert_requested(
            "get", "/v2/pageablemodels", {"page": {"token": "123"}, "foo": "bar"}, None
        )

        assert seen == ["pm_123", "pm_124", "pm_125", "pm_126"]

    def test_class_method_two_pages(self, request_mock):
        request_mock.stub_request(
            "get",
            "/v2/messaging_profiles",
            {"data": [{"id": "123", "record_type": "messaging_profile"}]},
        )

        seen = [item["id"] for item in telnyx.MessagingProfile.auto_paging_iter()]

        request_mock.assert_requested("get", "/v2/messaging_profiles", {})
        assert seen == ["123"]

    def test_class_method_two_pages_by_token(self, request_mock):
        request_mock.stub_request(
            "get",
            "/v2/messaging_profiles",
            {"data": [{"id": "123", "record_type": "messaging_profile"}]},
        )

        seen = [
            item["id"] for item in telnyx.MessagingProfile.auto_paging_iter_by_token()
        ]

        request_mock.assert_requested("get", "/v2/messaging_profiles", {})
        assert seen == ["123"]

    def test_previous_page(self, request_mock):
        url = "/v2/pageablemodels"
        lo = telnyx.ListObject.construct_from(
            self.pageable_model_response(["pm_125", "pm_126"], False, True, url),
            "mykey",
        )

        assert lo.page_number() == 2

        request_mock.stub_request(
            "get",
            "/v2/pageablemodels",
            self.pageable_model_response(["pm_123", "pm_124"], False, False, url),
        )

        lo = lo.previous_page()

        request_mock.assert_requested(
            "get", "/v2/pageablemodels", {"page": {"size": 20, "number": 1}}, None
        )

        assert lo.page_number() == 1



from __future__ import absolute_import, division, print_function

import pytest
import telnyx

TEST_RESOURCE_ID = "6a09cdc3-8948-47f0-aa62-74ac943d6c58"


class TestSIMCardGroups(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.SIMCardGroup.list()
        request_mock.assert_requested("get", "/v2/sim_card_groups")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.SIMCardGroup)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.SIMCardGroup.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/sim_card_groups/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.SIMCardGroup)

    def test_is_creatable(self, request_mock):
        resource = telnyx.SIMCardGroup.create(name="My Test Group")
        request_mock.assert_requested("post", "/v2/sim_card_groups")
        assert isinstance(resource, telnyx.SIMCardGroup)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_saveable(self, request_mock):
        sim_card_group = telnyx.SIMCardGroup.retrieve(TEST_RESOURCE_ID)
        sim_card_group.name = "My Test Group"
        resource = sim_card_group.save()
        request_mock.assert_requested(
            "patch", "/v2/sim_card_groups/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.SIMCardGroup)
        assert resource is sim_card_group

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_modifiable(self, request_mock):
        resource = telnyx.SIMCardGroup.modify(TEST_RESOURCE_ID, name="Updated Name")
        request_mock.assert_requested(
            "patch", "/v2/sim_card_groups/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.SIMCardGroup)

    def test_is_deletable(self, request_mock):
        resource = telnyx.SIMCardGroup.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/v2/sim_card_groups/%s" % TEST_RESOURCE_ID
        )



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("phone_number", operations=["list"])
@nested_resource_class_methods("short_code", operations=["list"])
@nested_resource_class_methods("metrics", operations=["list"])
class MessagingProfile(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "messaging_profile"

    def phone_numbers(self, **params):
        return MessagingProfile.list_phone_numbers(self.id, **params)

    def short_codes(self, **params):
        return MessagingProfile.list_short_codes(self.id, **params)

    def metrics(self, **params):
        return MessagingProfile.list_metrics(self.id, **params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("delete", path="actions/delete", operations=["delete"])
class CallRecording(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "call_recording"

    def delete(self, **params):
        return self.delete_delete(**params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import CreateableAPIResource, ListableAPIResource


class Message(CreateableAPIResource, ListableAPIResource):
    OBJECT_NAME = "message"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("refresh", path="actions/refresh", operations=["create"])
@nested_resource_class_methods("cancel", path="actions/cancel", operations=["create"])
class Fax(CreateableAPIResource, DeletableAPIResource, ListableAPIResource):
    OBJECT_NAME = "fax"

    @classmethod
    def class_url(cls):
        return "/v2/faxes"

    def refresh(self, **params):
        return self.create_refresh(**params)

    def cancel(self, **params):
        return self.create_cancel(**params)



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "6a09cdc3-8948-47f0-aa62-74ac943d6c58"


class TestNotificationChannel(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.NotificationChannel.list()
        request_mock.assert_requested("get", "/v2/notification_channels")
        assert isinstance(resources.data, list)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.NotificationChannel.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/notification_channels/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.NotificationChannel)

    @pytest.mark.skip(reason="Create mismatch")
    def test_is_creatable(self, request_mock):
        resource = telnyx.NotificationChannel.create()
        request_mock.assert_requested("post", "/v2/notification_channels")
        assert isinstance(resource, telnyx.NotificationChannel)

    @pytest.mark.skip(reason="Create mismatch")
    def test_is_modifiable(self, request_mock):
        resource = telnyx.NotificationChannel.modify(TEST_RESOURCE_ID, name="Test")
        request_mock.assert_requested(
            "patch", "/v2/notification_channels/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.NotificationChannel)

    @pytest.mark.skip(reason="Not released, no path yet")
    def test_is_deletable(self, request_mock):
        resource = telnyx.NotificationChannel.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/v2/notification_channels/%s" % TEST_RESOURCE_ID
        )



from __future__ import absolute_import, division, print_function

import io
import logging
import os
import re
import sys

import six
from six.moves.urllib.parse import parse_qsl

import telnyx

TELNYX_LOG = os.environ.get("TELNYX_LOG")

logger = logging.getLogger("telnyx")

__all__ = ["io", "parse_qsl", "utf8", "log_info", "log_debug", "logfmt"]

# Special valid characters in telnyx id's objects. Format: `=/$#`
telnyx_valid_id_parts = "=-"


def utf8(value):
    if six.PY2 and isinstance(value, six.text_type):
        return value.encode("utf-8")
    else:
        return value


def is_appengine_dev():
    return "APPENGINE_RUNTIME" in os.environ and "Dev" in os.environ.get(
        "SERVER_SOFTWARE", ""
    )


def _console_log_level():
    if telnyx.log in ["debug", "info"]:
        return telnyx.log
    elif TELNYX_LOG in ["debug", "info"]:
        return TELNYX_LOG
    else:
        return None


# Rewrites reserved word keyword arguments
# This makes Message.create(to="", from_="", text="") possible
def rewrite_reserved_words(kwargs, reverse=False):
    reserved = [("from_", "from")]

    for i in reserved:
        if not reverse:
            original, replacement = i
        else:
            replacement, original = i

        if kwargs.get(original, None) is not None:
            kwargs[replacement] = kwargs.pop(original)

    return kwargs


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == "debug":
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ["debug", "info"]:
        print(msg, file=sys.stderr)
    logger.info(msg)


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if six.PY3 and hasattr(val, "decode"):
            val = val.decode("utf-8")
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, six.string_types):
            val = six.text_type(val)
        if re.search(r"\s", val):
            val = repr(val)
        # key should already be a string
        if re.search(r"\s", key):
            key = repr(key)
        return "{key}={val}".format(key=key, val=val)

    return " ".join([fmt(key, val) for key, val in sorted(props.items())])


OBJECT_CLASSES = {}


def load_object_classes():
    # This is here to avoid a circular dependency
    from telnyx import api_resources

    global OBJECT_CLASSES

    OBJECT_CLASSES = {
        # business objects
        api_resources.APIKey.OBJECT_NAME: api_resources.APIKey,
        api_resources.AccessControlIP.OBJECT_NAME: api_resources.AccessControlIP,
        api_resources.AccessIPAddress.OBJECT_NAME: api_resources.AccessIPAddress,
        api_resources.AccessToken.OBJECT_NAME: api_resources.AccessToken,
        api_resources.Address.OBJECT_NAME: api_resources.Address,
        api_resources.AdvancedOptinoptout.OBJECT_NAME: api_resources.AdvancedOptinoptout,
        api_resources.AI.OBJECT_NAME: api_resources.AI,
        api_resources.AuthenticationProvider.OBJECT_NAME: api_resources.AuthenticationProvider,
        api_resources.Autorechargepreference.OBJECT_NAME: api_resources.Autorechargepreference,
        api_resources.AvailablePhoneNumber.OBJECT_NAME: api_resources.AvailablePhoneNumber,
        api_resources.Balance.OBJECT_NAME: api_resources.Balance,
        api_resources.Billing.OBJECT_NAME: api_resources.Billing,
        api_resources.BillingGroup.OBJECT_NAME: api_resources.BillingGroup,
        api_resources.Brand.OBJECT_NAME: api_resources.Brand,
        api_resources.Bucket.OBJECT_NAME: api_resources.Bucket,
        api_resources.BucketUsage.OBJECT_NAME: api_resources.BucketUsage,
        api_resources.BulkCreation.OBJECT_NAME: api_resources.BulkCreation,
        api_resources.BulkCredential.OBJECT_NAME: api_resources.BulkCredential,
        api_resources.BulkPhoneNumberCampaign.OBJECT_NAME: api_resources.BulkPhoneNumberCampaign,
        api_resources.BulkPhoneNumberOperation.OBJECT_NAME: api_resources.BulkPhoneNumberOperation,
        api_resources.BulkSoleProprietorCreation.OBJECT_NAME: api_resources.BulkSoleProprietorCreation,
        api_resources.BusinessIdentity.OBJECT_NAME: api_resources.BusinessIdentity,
        api_resources.Call.OBJECT_NAME: api_resources.Call,
        api_resources.CallControlApplication.OBJECT_NAME: api_resources.CallControlApplication,
        api_resources.CallInformation.OBJECT_NAME: api_resources.CallInformation,
        api_resources.CallRecording.OBJECT_NAME: api_resources.CallRecording,
        api_resources.Campaign.OBJECT_NAME: api_resources.Campaign,
        api_resources.CdrUsageReport.OBJECT_NAME: api_resources.CdrUsageReport,
        api_resources.Comment.OBJECT_NAME: api_resources.Comment,
        api_resources.Conference.OBJECT_NAME: api_resources.Conference,
        api_resources.Connection.OBJECT_NAME: api_resources.Connection,
        api_resources.CredentialConnection.OBJECT_NAME: api_resources.CredentialConnection,
        api_resources.CsvDownload.OBJECT_NAME: api_resources.CsvDownload,
        api_resources.CustomStorageCredential.OBJECT_NAME: api_resources.CustomStorageCredential,
        api_resources.CustomerServiceRecord.OBJECT_NAME: api_resources.CustomerServiceRecord,
        api_resources.Debugging.OBJECT_NAME: api_resources.Debugging,
        api_resources.DetailRecord.OBJECT_NAME: api_resources.DetailRecord,
        api_resources.DetailRecordsReport.OBJECT_NAME: api_resources.DetailRecordsReport,
        api_resources.DialogflowIntegration.OBJECT_NAME: api_resources.DialogflowIntegration,
        api_resources.Document.OBJECT_NAME: api_resources.Document,
        api_resources.Document.OBJECT_NAME: api_resources.DocumentLink,
        api_resources.DynamicEmergencyAddress.OBJECT_NAME: api_resources.DynamicEmergencyAddress,
        api_resources.DynamicEmergencyEndpoint.OBJECT_NAME: api_resources.DynamicEmergencyEndpoint,
        api_resources.Enum.OBJECT_NAME: api_resources.Enum,
        api_resources.Event.OBJECT_NAME: api_resources.Event,
        api_resources.ExternalConnection.OBJECT_NAME: api_resources.ExternalConnection,
        api_resources.Fax.OBJECT_NAME: api_resources.Fax,
        api_resources.FaxApplication.OBJECT_NAME: api_resources.FaxApplication,
        api_resources.FQDN.OBJECT_NAME: api_resources.FQDN,
        api_resources.FQDNConnection.OBJECT_NAME: api_resources.FQDNConnection,
        api_resources.IP.OBJECT_NAME: api_resources.IP,
        api_resources.IPAddress.OBJECT_NAME: api_resources.IPAddress,
        api_resources.IPConnection.OBJECT_NAME: api_resources.IPConnection,
        api_resources.IPRange.OBJECT_NAME: api_resources.IPRange,
        api_resources.InboundChannel.OBJECT_NAME: api_resources.InboundChannel,
        api_resources.InventoryCoverage.OBJECT_NAME: api_resources.InventoryCoverage,
        api_resources.InventoryLevel.OBJECT_NAME: api_resources.InventoryLevel,
        api_resources.ManagedAccount.OBJECT_NAME: api_resources.ManagedAccount,
        api_resources.MdrDetailReport.OBJECT_NAME: api_resources.MdrDetailReport,
        api_resources.MdrUsageReport.OBJECT_NAME: api_resources.MdrUsageReport,
        api_resources.Media.OBJECT_NAME: api_resources.Media,
        api_resources.MediaStorageApi.OBJECT_NAME: api_resources.MediaStorageApi,
        api_resources.Message.OBJECT_NAME: api_resources.Message,
        api_resources.MessagingHostedNumber.OBJECT_NAME: api_resources.MessagingHostedNumber,
        api_resources.MessagingHostedNumberOrder.OBJECT_NAME: api_resources.MessagingHostedNumberOrder,
        api_resources.MessagingPhoneNumber.OBJECT_NAME: api_resources.MessagingPhoneNumber,
        api_resources.MessagingProfile.OBJECT_NAME: api_resources.MessagingProfile,
        api_resources.MessagingTollfreeVerification.OBJECT_NAME: api_resources.MessagingTollfreeVerification,
        api_resources.MessagingUrlDomain.OBJECT_NAME: api_resources.MessagingUrlDomain,
        api_resources.MobileNetworkOperator.OBJECT_NAME: api_resources.MobileNetworkOperator,
        api_resources.Network.OBJECT_NAME: api_resources.Network,
        api_resources.Notification.OBJECT_NAME: api_resources.Notification,
        api_resources.NotificationChannel.OBJECT_NAME: api_resources.NotificationChannel,
        api_resources.NotificationEvent.OBJECT_NAME: api_resources.NotificationEvent,
        api_resources.NotificationEventCondition.OBJECT_NAME: api_resources.NotificationEventCondition,
        api_resources.NotificationProfile.OBJECT_NAME: api_resources.NotificationProfile,
        api_resources.NotificationSetting.OBJECT_NAME: api_resources.NotificationSetting,
        api_resources.NumberConfiguration.OBJECT_NAME: api_resources.NumberConfiguration,
        api_resources.NumberLookup.OBJECT_NAME: api_resources.NumberLookup,
        api_resources.NumberOrder.OBJECT_NAME: api_resources.NumberOrder,
        api_resources.NumberOrderPhoneNumber.OBJECT_NAME: api_resources.NumberOrderPhoneNumber,
        api_resources.NumberPortout.OBJECT_NAME: api_resources.NumberPortout,
        api_resources.NumberReservation.OBJECT_NAME: api_resources.NumberReservation,
        api_resources.NumbersFeature.OBJECT_NAME: api_resources.NumbersFeature,
        api_resources.OtaUpdate.OBJECT_NAME: api_resources.OtaUpdate,
        api_resources.OutboundVoiceProfile.OBJECT_NAME: api_resources.OutboundVoiceProfile,
        api_resources.PhoneNumber.OBJECT_NAME: api_resources.PhoneNumber,
        api_resources.PhoneNumber.OBJECT_NAME: api_resources.MessagingSettings,
        api_resources.PhoneNumber.OBJECT_NAME: api_resources.VoiceSettings,
        api_resources.PhoneNumberBlockOrder.OBJECT_NAME: api_resources.PhoneNumberBlockOrder,
        api_resources.PhoneNumberBlocksBackgroundJob.OBJECT_NAME: api_resources.PhoneNumberBlocksBackgroundJob,
        api_resources.PhoneNumberByProfile.OBJECT_NAME: api_resources.PhoneNumberByProfile,
        api_resources.PhoneNumberCampaign.OBJECT_NAME: api_resources.PhoneNumberCampaign,
        api_resources.PhoneNumberConfiguration.OBJECT_NAME: api_resources.PhoneNumberConfiguration,
        api_resources.PhoneNumberJob.OBJECT_NAME: api_resources.PhoneNumberJob,
        api_resources.PhoneNumberOrder.OBJECT_NAME: api_resources.PhoneNumberOrder,
        api_resources.PhoneNumberOrderDocument.OBJECT_NAME: api_resources.PhoneNumberOrderDocument,
        api_resources.PhoneNumberPorting.OBJECT_NAME: api_resources.PhoneNumberPorting,
        api_resources.PhoneNumberRegulatoryRequirement.OBJECT_NAME: api_resources.PhoneNumberRegulatoryRequirement,
        api_resources.PhoneNumberReservation.OBJECT_NAME: api_resources.PhoneNumberReservation,
        api_resources.PhoneNumberSearch.OBJECT_NAME: api_resources.PhoneNumberSearch,
        api_resources.PortOut.OBJECT_NAME: api_resources.PortOut,
        api_resources.PortabilityCheck.OBJECT_NAME: api_resources.PortabilityCheck,
        api_resources.PortingOrder.OBJECT_NAME: api_resources.PortingOrder,
        api_resources.PortingOrder.OBJECT_NAME: api_resources.PortingPhoneNumber,
        api_resources.PresignedObjectUrl.OBJECT_NAME: api_resources.PresignedObjectUrl,
        api_resources.PrivateWirelessGateway.OBJECT_NAME: api_resources.PrivateWirelessGateway,
        api_resources.PublicInternetGateway.OBJECT_NAME: api_resources.PublicInternetGateway,
        api_resources.PushCredential.OBJECT_NAME: api_resources.PushCredential,
        api_resources.Queue.OBJECT_NAME: api_resources.Queue,
        api_resources.Queue.OBJECT_NAME: api_resources.QueueCall,
        api_resources.QueueCommand.OBJECT_NAME: api_resources.QueueCommand,
        api_resources.RecordingsCommand.OBJECT_NAME: api_resources.RecordingsCommand,
        api_resources.Region.OBJECT_NAME: api_resources.Region,
        api_resources.RegisterCall.OBJECT_NAME: api_resources.RegisterCall,
        api_resources.Report.OBJECT_NAME: api_resources.Report,
        api_resources.Reporting.OBJECT_NAME: api_resources.Reporting,
        api_resources.Requirement.OBJECT_NAME: api_resources.Requirement,
        api_resources.RequirementType.OBJECT_NAME: api_resources.RequirementType,
        api_resources.Room.OBJECT_NAME: api_resources.Room,
        api_resources.RoomComposition.OBJECT_NAME: api_resources.RoomComposition,
        api_resources.RoomParticipant.OBJECT_NAME: api_resources.RoomParticipant,
        api_resources.RoomRecording.OBJECT_NAME: api_resources.RoomRecording,
        api_resources.RoomSession.OBJECT_NAME: api_resources.RoomSession,
        api_resources.RoomsClientToken.OBJECT_NAME: api_resources.RoomsClientToken,
        api_resources.SIMCard.OBJECT_NAME: api_resources.SIMCard,
        api_resources.SIMCardAction.OBJECT_NAME: api_resources.SIMCardAction,
        api_resources.SIMCardGroup.OBJECT_NAME: api_resources.SIMCardGroup,
        api_resources.SIMCardOrder.OBJECT_NAME: api_resources.SIMCardOrder,
        api_resources.SIMCardOrder.OBJECT_NAME: api_resources.SIMCardOrderPreview,
        api_resources.SharedCampaign.OBJECT_NAME: api_resources.SharedCampaign,
        api_resources.ShortCode.OBJECT_NAME: api_resources.ShortCode,
        api_resources.SslCertificate.OBJECT_NAME: api_resources.SslCertificate,
        api_resources.SubNumberOrder.OBJECT_NAME: api_resources.SubNumberOrder,
        api_resources.TelephonyCredential.OBJECT_NAME: api_resources.TelephonyCredential,
        api_resources.TexmlApplication.OBJECT_NAME: api_resources.TexmlApplication,
        api_resources.TexmlRestCommand.OBJECT_NAME: api_resources.TexmlRestCommand,
        api_resources.Verification.OBJECT_NAME: api_resources.Verification,
        api_resources.VerifiedCallsDisplayProfile.OBJECT_NAME: api_resources.VerifiedCallsDisplayProfile,
        api_resources.VerifiedNumber.OBJECT_NAME: api_resources.VerifiedNumber,
        api_resources.Verify.OBJECT_NAME: api_resources.Verify,
        api_resources.VerifyProfile.OBJECT_NAME: api_resources.VerifyProfile,
        api_resources.VirtualCrossConnect.OBJECT_NAME: api_resources.VirtualCrossConnect,
        api_resources.WdrDetailReport.OBJECT_NAME: api_resources.WdrDetailReport,
        api_resources.Webhook.OBJECT_NAME: api_resources.Webhook,
        api_resources.WebhookDeliveries.OBJECT_NAME: api_resources.WebhookDeliveries,
        api_resources.WireguardInterface.OBJECT_NAME: api_resources.WireguardInterface,
        api_resources.WirelessDetailRecordsReports.OBJECT_NAME: api_resources.WirelessDetailRecordsReports,
    }


def convert_to_telnyx_object(resp, api_key=None):
    global OBJECT_CLASSES

    if len(OBJECT_CLASSES) == 0:
        load_object_classes()
    types = OBJECT_CLASSES.copy()

    # If we get a TelnyxResponse, we'll want to return a
    # TelnyxObject with the last_response field filled out with
    # the raw API response information
    telnyx_response = None

    if isinstance(resp, telnyx.telnyx_response.TelnyxResponse):
        telnyx_response = resp
        resp = telnyx_response.data

    if isinstance(resp, list):
        return [convert_to_telnyx_object(i, api_key) for i in resp]
    elif isinstance(resp, dict) and not isinstance(
        resp, telnyx.telnyx_object.TelnyxObject
    ):
        resp = resp.copy()
        resp = rewrite_reserved_words(resp, reverse=True)
        data = resp.get("data", None)
        if data:
            if isinstance(data, list):
                return telnyx.api_resources.ListObject.construct_from(
                    resp, api_key, last_response=telnyx_response
                )

            record_type = data.get("record_type", None)
            if record_type:
                klass = types.get(record_type, telnyx.telnyx_object.TelnyxObject)
                return klass.construct_from(
                    data, api_key, last_response=telnyx_response
                )

        record_type = resp.get("record_type", None)
        if record_type:
            klass = types.get(record_type, telnyx.telnyx_object.TelnyxObject)
            return klass.construct_from(resp, api_key, last_response=telnyx_response)
        else:
            klass = telnyx.telnyx_object.TelnyxObject
            return klass.construct_from(resp, api_key, last_response=telnyx_response)
    else:
        return resp



from __future__ import absolute_import, division, print_function

import telnyx


class TestListableAPIResource(object):
    class MyListable(telnyx.api_resources.abstract.ListableAPIResource):
        OBJECT_NAME = "mylistable"

    def test_all(self, request_mock):
        request_mock.stub_request(
            "get",
            "/v2/mylistables",
            {
                "data": [
                    {"record_type": "messaging_profile", "name": "jose"},
                    {"record_type": "messaging_profile", "name": "curly"},
                ]
            },
            rheaders={"request-id": "req_id"},
        )

        res = self.MyListable.list()
        request_mock.assert_requested("get", "/v2/mylistables", {})
        assert len(res.data) == 2
        assert all(isinstance(obj, telnyx.MessagingProfile) for obj in res.data)
        assert res.data[0].name == "jose"
        assert res.data[1].name == "curly"

        assert res.last_response is not None
        assert res.last_response.request_id == "req_id"



from __future__ import absolute_import, division, print_function

import base64
import json
import time

from nacl.encoding import Base64Encoder
from nacl.signing import VerifyKey

import telnyx
from telnyx import error


class Webhook(object):
    DEFAULT_TOLERANCE = 300

    @staticmethod
    def construct_event(
        payload, signature_header, timestamp, tolerance=DEFAULT_TOLERANCE, api_key=None
    ):
        if WebhookSignature.verify(payload, signature_header, timestamp, tolerance):
            if api_key is None:
                api_key = telnyx.api_key

            if hasattr(payload, "decode"):
                payload = payload.decode("utf-8")

            data = json.loads(payload)
            event = telnyx.Event.construct_from(data, api_key)

            return event


class WebhookSignature(object):
    @classmethod
    def verify(cls, payload, signature, timestamp, tolerance=None):
        if hasattr(timestamp, "encode"):
            timestamp = timestamp.encode("utf-8")

        if hasattr(payload, "encode"):
            payload = payload.encode("utf-8")

        signed_payload = timestamp + b"|" + payload
        public_key = telnyx.public_key

        if not public_key:
            raise error.SignatureVerificationError(
                "Public key not set", signature, payload
            )

        # verify the data
        key = VerifyKey(public_key, encoder=Base64Encoder)

        if not key.verify(signed_payload, signature=base64.b64decode(signature)):
            raise error.SignatureVerificationError(
                "Signature is invalid and does not match the payload",
                signature,
                payload,
            )

        if tolerance and int(timestamp) < time.time() - tolerance:
            raise error.SignatureVerificationError(
                "Timestamp outside the tolerance zone (%s)" % timestamp,
                signature,
                payload,
            )

        return True



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "6a09cdc3-8948-47f0-aa62-74ac943d6c58"


class TestSIMCard(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.SIMCard.list()
        request_mock.assert_requested("get", "/v2/sim_cards")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.SIMCard)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/sim_cards/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.SIMCard)

    def test_is_saveable(self, request_mock):
        sim_card = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        sim_card.tags = ["foo", "bar"]
        resource = sim_card.save()
        request_mock.assert_requested("patch", "/v2/sim_cards/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.SIMCard)
        assert resource is sim_card

    def test_is_modifiable(self, request_mock):
        resource = telnyx.SIMCard.modify(TEST_RESOURCE_ID, tags=["foo", "bar"])
        request_mock.assert_requested("patch", "/v2/sim_cards/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.SIMCard)

    def test_is_deletabale(self, request_mock):
        sim_card = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        sim_card.delete()
        request_mock.assert_requested("delete", "/v2/sim_cards/%s" % TEST_RESOURCE_ID)
        assert isinstance(sim_card, telnyx.SIMCard)

    @pytest.mark.skip(reason="Async, need to wait for promise")
    def test_enable(self, request_mock):
        resource = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        resource.create_enable()
        request_mock.assert_requested(
            "post", "/v2/sim_cards/%s/actions/enable" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Async, need to wait for promise")
    def test_disable(self, request_mock):
        resource = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        resource.create_disable()
        request_mock.assert_requested(
            "post", "/v2/sim_cards/%s/actions/disable" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Async, need to wait for promise")
    def test_set_standby(self, request_mock):
        telnyx.SIMCard.create_set_standby(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v2/sim_cards/%s/actions/set_standby" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Async, need to wait for promise")
    def test_remove_public_ip(self, request_mock):
        telnyx.SIMCard.create_remove_public_ip(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v2/sim_cards/%s/actions/remove_public_ip" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Async, need to wait for promise")
    def test_set_public_ip(self, request_mock):
        telnyx.SIMCard.create_set_public_ip(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v2/sim_cards/%s/actions/set_public_ip" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="deprecated?")
    def test_wireless_connectivity_logs(self, request_mock):
        source = telnyx.SIMCard.retrieve(TEST_RESOURCE_ID)
        source.list_wireless_logs()

    def test_register(self, request_mock):
        telnyx.SIMCard.register(registration_codes=["foo", "bar"])
        request_mock.assert_requested("post", "/v2/actions/register/sim_cards")



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import CreateableAPIResource, ListableAPIResource


class Comment(ListableAPIResource, CreateableAPIResource):
    OBJECT_NAME = "comment"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("sms", path="verifications/sms", operations=["create"])
@nested_resource_class_methods("call", path="verifications/call", operations=["create"])
@nested_resource_class_methods("flashcall", path="verifications/flashcall", operations=["create"])
@nested_resource_class_methods("verify", path="verifications/{verification_id}/actions/verify", operations=["create"])
class Verify(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "verify"

    def create_verification_sms(self, **params):
        return self.create_sms(**params)

    def create_verification_call(self, **params):
        return self.create_call(**params)

    def verify_verification_code_by_id(self, verification_id, **params):
        return self.create_verify(verification_id=verification_id, **params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "comments", path="comments", operations=["create", "list"], pluralize_path=False
)
@nested_resource_class_methods(
    "supporting_documents",
    path="supporting_documents",
    operations=["create", "list"],
    pluralize_path=False,
)
class PortOut(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "portout"

    def list_comments(self, **params):
        return PortOut.list_comments(self.id, **params)

    def create_comments(self, **params):
        return PortOut.create_comments(self, **params)

    def list_supporting_documents(self, **params):
        return PortOut.list_supporting_documents(self.id, **params)

    def create_supporting_documents(self, **params):
        return PortOut.create_supporting_documents(self, **params)



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "35146afd-df93-4963-b1e9-1a085e2ae874"


class TestAuthenticationProvider(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.AuthenticationProvider.list()
        request_mock.assert_requested("get", "/v2/authentication_providers")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.AuthenticationProvider)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_retrievable(self, request_mock):
        resource = telnyx.AuthenticationProvider.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/authentication_providers/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.AuthenticationProvider)

    @pytest.mark.skip(reason="Formatting for setting param is off")
    def test_is_creatable(self, request_mock):
        resource = telnyx.AuthenticationProvider.create(
            name="Okta", setting=[], short_name="myorg"
        )
        request_mock.assert_requested("post", "/v2/authentication_providers")
        assert isinstance(resource, telnyx.AuthenticationProvider)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_saveable(self, request_mock):
        authentication_provider = telnyx.AuthenticationProvider.retrieve(
            TEST_RESOURCE_ID
        )
        authentication_provider.name = "value"
        resource = authentication_provider.save()
        request_mock.assert_requested(
            "patch", "/v2/authentication_providers/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.AuthenticationProvider)
        assert resource is authentication_provider

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_modifiable(self, request_mock):
        resource = telnyx.AuthenticationProvider.modify(TEST_RESOURCE_ID, active=False)
        request_mock.assert_requested(
            "patch", "/v2/authentication_providers/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.AuthenticationProvider)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_deletable(self, request_mock):
        resource = telnyx.AuthenticationProvider.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/v2/authentication_providers/%s" % TEST_RESOURCE_ID
        )



from __future__ import absolute_import, division, print_function

import atexit
import os
import sys

import pytest
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlopen

from tests.request_mock import RequestMock
from tests.telnyx_mock import TelnyxMock

import telnyx

# When changing this number, don't forget to change it in `.travis.yml` too.
MOCK_MINIMUM_VERSION = "0.8.10"

# Starts telnyx-mock if an OpenAPI spec override is found in `openapi/`, and
# otherwise fall back to `TELNYX_MOCK_PORT` or 12111.
if TelnyxMock.start():
    MOCK_PORT = TelnyxMock.port()
else:
    MOCK_PORT = os.environ.get("TELNYX_MOCK_PORT", 12111)


@atexit.register
def stop_telnyx_mock():
    TelnyxMock.stop()


try:
    resp = urlopen("http://localhost:%s/" % MOCK_PORT)
    info = resp.info()
except HTTPError as e:
    info = e.info()
except Exception:
    sys.exit(
        "Couldn't reach telnyx-mock at `localhost:%s`. Is "
        "it running? Please see README for setup instructions." % MOCK_PORT
    )

"""
version = info.get("Telnyx-Mock-Version")
if version != "master" and StrictVersion(version) < StrictVersion(MOCK_MINIMUM_VERSION):
    sys.exit(
        "Your version of telnyx-mock (%s) is too old. The minimum "
        "version to run this test suite is %s. Please "
        "see its repository for upgrade instructions." % (version, MOCK_MINIMUM_VERSION)
    )
"""


@pytest.fixture(autouse=True)
def setup_telnyx():
    orig_attrs = {
        "api_base": telnyx.api_base,
        "api_key": telnyx.api_key,
        "default_http_client": telnyx.default_http_client,
    }
    http_client = telnyx.http_client.new_default_http_client()
    telnyx.api_base = "http://localhost:%s" % MOCK_PORT
    telnyx.api_key = "KEY123"
    telnyx.default_http_client = http_client
    yield
    http_client.close()
    telnyx.api_base = orig_attrs["api_base"]
    telnyx.api_key = orig_attrs["api_key"]
    telnyx.default_http_client = orig_attrs["default_http_client"]


@pytest.fixture
def request_mock(mocker):
    return RequestMock(mocker)



from __future__ import absolute_import, division, print_function

import telnyx


class TestBalance(object):
    def test_is_retrievable(self, request_mock):
        resource = telnyx.Balance.retrieve()
        request_mock.assert_requested("get", "/v2/balance")
        assert isinstance(resource, telnyx.Balance)



from __future__ import absolute_import, division, print_function

import os

from telnyx.api_resources import *  # noqa
from telnyx.webhook import Webhook, WebhookSignature  # noqa

# Telnyx Python bindings
# API docs at http://developers.telnyx.com

# Configuration variables

api_key = os.environ.get("TELNYX_API_KEY")
api_base = os.environ.get("TELNYX_API_BASE", "https://api.telnyx.com")
verify_ssl_certs = True
proxy = None
default_http_client = None
app_info = None
max_network_retries = 0

public_key = os.environ.get("TELNYX_PUBLIC_KEY")

# Set to either 'debug' or 'info', controls console logging
log = None


__version__ = "2.1.4"


# Sets some basic information about the running application that's sent along
# with API requests. Useful for plugin authors to identify their plugin when
# communicating with Telnyx.
#
# Takes a name and optional version and plugin URL.
def set_app_info(name, url=None, version=None):
    global app_info
    app_info = {"name": name, "url": url, "version": version}



from __future__ import absolute_import, division, print_function

import pytest

import telnyx


class TestPublicKey(object):
    @pytest.mark.skip(reason="Unsure about this endpoint, might be deprecated")
    def test_is_retrievable(self, request_mock):
        resource = telnyx.PublicKey.retrieve()
        request_mock.assert_requested("get", "/v2/public_key")
        assert isinstance(resource, telnyx.PublicKey)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)


class AuthenticationProvider(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "authentication_provider"



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import quote_plus

from telnyx import api_requestor, util


def nested_resource_class_methods(
    resource, path=None, operations=None, pluralize_path=True
):
    def make_path(v):
        if pluralize_path:
            return "%ss" % v
        else:
            return v

    if path is None:
        path = make_path(resource)
    if operations is None:
        raise ValueError("operations list required")

    def wrapper(cls):
        def nested_resource_url(cls, id, nested_id=None):
            parts = []
            if not path.startswith("/"):
                parts.append(cls.class_url())
            if id is not None and "phone_number" not in path:
                parts.append(quote_plus(id, safe=util.telnyx_valid_id_parts))
            if id is not None:
                if "phone_number" in path:
                    parts.append(path.format(phone_number=quote_plus(id, safe=util.telnyx_valid_id_parts + '+')))
                elif "verification_id" in path:
                    parts.append(path.format(verification_id=quote_plus(id, safe=util.telnyx_valid_id_parts)))
                else:
                    parts.append(path)
            else:
                parts.append(path)
            if nested_id is not None:
                parts.append(quote_plus(nested_id, safe=util.telnyx_valid_id_parts))
            return "/".join(parts)

        resource_url_method = "%s_url" % make_path(resource)
        setattr(cls, resource_url_method, classmethod(nested_resource_url))

        def nested_resource_request(cls, method, url, api_key=None, **params):
            requestor = api_requestor.APIRequestor(api_key)
            params = util.rewrite_reserved_words(params)
            response, api_key = requestor.request(method, url, params)
            return util.convert_to_telnyx_object(response, api_key)

        resource_request_method = "%s_request" % make_path(resource)
        setattr(cls, resource_request_method, classmethod(nested_resource_request))

        for operation in operations:
            if operation == "create":

                def create_nested_resource(cls, id, **params):
                    url = getattr(cls, resource_url_method)(id)
                    return getattr(cls, resource_request_method)("post", url, **params)

                create_method = "create_%s" % resource
                setattr(cls, create_method, classmethod(create_nested_resource))

            elif operation == "retrieve":

                def retrieve_nested_resource(cls, id, nested_id=None, **params):
                    url = getattr(cls, resource_url_method)(id, nested_id=nested_id)
                    return getattr(cls, resource_request_method)("get", url, **params)

                retrieve_method = "retrieve_%s" % resource
                setattr(cls, retrieve_method, classmethod(retrieve_nested_resource))

            elif operation == "update":

                def modify_nested_resource(cls, id, nested_id, **params):
                    url = getattr(cls, resource_url_method)(id, nested_id)
                    return getattr(cls, resource_request_method)("post", url, **params)

                modify_method = "modify_%s" % resource
                setattr(cls, modify_method, classmethod(modify_nested_resource))

            elif operation == "delete":

                def delete_nested_resource(cls, id, nested_id, **params):
                    url = getattr(cls, resource_url_method)(id, nested_id)
                    return getattr(cls, resource_request_method)(
                        "delete", url, **params
                    )

                delete_method = "delete_%s" % resource
                setattr(cls, delete_method, classmethod(delete_nested_resource))

            elif operation == "list":

                def list_nested_resources(cls, id, **params):
                    url = getattr(cls, resource_url_method)(id)
                    return getattr(cls, resource_request_method)("get", url, **params)

                list_method = "list_%s" % make_path(resource)
                setattr(cls, list_method, classmethod(list_nested_resources))

            elif operation == "put":

                def update_nested_resource(cls, id, nested_id, **params):
                    url = getattr(cls, resource_url_method)(id, nested_id)
                    return getattr(cls, resource_request_method)("put", url, **params)

                update_method = "update_%s" % resource
                setattr(cls, update_method, classmethod(update_nested_resource))

            else:
                raise ValueError("Unknown operation: %s" % operation)

        return cls

    return wrapper



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
)


class DynamicEmergencyEndpoint(
    CreateableAPIResource, DeletableAPIResource, ListableAPIResource
):
    OBJECT_NAME = "dynamic_emergency_endpoint"



from __future__ import absolute_import, division, print_function

from telnyx.telnyx_object import TelnyxObject


class NumberOrderPhoneNumber(TelnyxObject):
    OBJECT_NAME = "number_order_phone_number"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "transcription_stop", path="actions/transcription_stop", operations=["create"]
)
@nested_resource_class_methods(
    "fork_start", path="actions/fork_start", operations=["create"]
)
@nested_resource_class_methods(
    "suppression_start", path="actions/suppression_start", operations=["create"]
)
@nested_resource_class_methods(
    "transfer", path="actions/transfer", operations=["create"]
)
@nested_resource_class_methods("answer", path="actions/answer", operations=["create"])
@nested_resource_class_methods(
    "playback_stop", path="actions/playback_stop", operations=["create"]
)
@nested_resource_class_methods(
    "suppression_stop", path="actions/suppression_stop", operations=["create"]
)
@nested_resource_class_methods(
    "gather_stop", path="actions/gather_stop", operations=["create"]
)
@nested_resource_class_methods("hangup", path="actions/hangup", operations=["create"])
@nested_resource_class_methods("refer", path="actions/refer", operations=["create"])
@nested_resource_class_methods(
    "record_resume", path="actions/record_resume", operations=["create"]
)
@nested_resource_class_methods(
    "gather_using_audio", path="actions/gather_using_audio", operations=["create"]
)
@nested_resource_class_methods(
    "send_dtmf", path="actions/send_dtmf", operations=["create"]
)
@nested_resource_class_methods("gather", path="actions/gather", operations=["create"])
@nested_resource_class_methods(
    "leave_queue", path="actions/leave_queue", operations=["create"]
)
@nested_resource_class_methods(
    "record_stop", path="actions/record_stop", operations=["create"]
)
@nested_resource_class_methods(
    "record_pause", path="actions/record_pause", operations=["create"]
)
@nested_resource_class_methods(
    "playback_start", path="actions/playback_start", operations=["create"]
)
@nested_resource_class_methods(
    "record_start", path="actions/record_start", operations=["create"]
)
@nested_resource_class_methods("bridge", path="actions/bridge", operations=["create"])
@nested_resource_class_methods(
    "streaming_start", path="actions/streaming_start", operations=["create"]
)
@nested_resource_class_methods("enqueue", path="actions/enqueue", operations=["create"])
@nested_resource_class_methods("speak", path="actions/speak", operations=["create"])
@nested_resource_class_methods(
    "fork_stop", path="actions/fork_stop", operations=["create"]
)
@nested_resource_class_methods("reject", path="actions/reject", operations=["create"])
@nested_resource_class_methods(
    "transcription_start", path="actions/transcription_start", operations=["create"]
)
@nested_resource_class_methods(
    "client_state_update", path="actions/client_state_update", operations=["put"]
)
@nested_resource_class_methods(
    "gather_using_speak", path="actions/gather_using_speak", operations=["create"]
)
@nested_resource_class_methods(
    "streaming_stop", path="actions/streaming_stop", operations=["create"]
)
class Call(CreateableAPIResource):
    OBJECT_NAME = "call"

    def transcription_stop(self, **params):
        return Call.create_transcription_stop(self.call_control_id, **params)

    def fork_start(self, **params):
        return Call.create_fork_start(self.call_control_id, **params)

    def suppression_start(self, **params):
        return Call.create_suppression_start(self.call_control_id, **params)

    def transfer(self, **params):
        return Call.create_transfer(self.call_control_id, **params)

    def answer(self, **params):
        return Call.create_answer(self.call_control_id, **params)

    def playback_stop(self, **params):
        return Call.create_playback_stop(self.call_control_id, **params)

    def suppression_stop(self, **params):
        return Call.create_suppression_stop(self.call_control_id, **params)

    def gather_stop(self, **params):
        return Call.create_gather_stop(self.call_control_id, **params)

    def hangup(self, **params):
        return Call.create_hangup(self.call_control_id, **params)

    def refer(self, **params):
        return Call.create_refer(self.call_control_id, **params)

    def record_resume(self, **params):
        return Call.create_record_resume(self.call_control_id, **params)

    def gather_using_audio(self, **params):
        return Call.create_gather_using_audio(self.call_control_id, **params)

    def send_dtmf(self, **params):
        return Call.create_send_dtmf(self.call_control_id, **params)

    def gather(self, **params):
        return Call.create_gather(self.call_control_id, **params)

    def leave_queue(self, **params):
        return Call.create_leave_queue(self.call_control_id, **params)

    def record_stop(self, **params):
        return Call.create_record_stop(self.call_control_id, **params)

    def record_pause(self, **params):
        return Call.create_record_pause(self.call_control_id, **params)

    def playback_start(self, **params):
        return Call.create_playback_start(self.call_control_id, **params)

    def record_start(self, **params):
        return Call.create_record_start(self.call_control_id, **params)

    def bridge(self, **params):
        return Call.create_bridge(self.call_control_id, **params)

    def streaming_start(self, **params):
        return Call.create_streaming_start(self.call_control_id, **params)

    def enqueue(self, queue_name, **params):
        params["queue_name"] = queue_name
        return Call.create_enqueue(self.call_control_id, **params)

    def speak(self, **params):
        return Call.create_speak(self.call_control_id, **params)

    def fork_stop(self, **params):
        return Call.create_fork_stop(self.call_control_id, **params)

    def reject(self, **params):
        return Call.create_reject(self.call_control_id, **params)

    def transcription_start(self, **params):
        return Call.create_transcription_start(self.call_control_id, **params)

    def client_state_update(self, **params):
        return Call.put_client_state_update(self.call_control_id, **params)

    def gather_using_speak(self, **params):
        return Call.create_gather_using_speak(self.call_control_id, **params)

    def streaming_stop(self, **params):
        return Call.create_streaming_stop(self.call_control_id, **params)



from __future__ import absolute_import, division, print_function

import pytest

import telnyx
from telnyx import error

TEST_RESOURCE_ID = "1293384261075731499"


class TestPhoneNumber(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.PhoneNumber.list()
        request_mock.assert_requested("get", "/v2/phone_numbers")
        assert isinstance(resources.data, list)

    def test_is_retrievable(self, request_mock):
        telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/phone_numbers/%s" % TEST_RESOURCE_ID)

    def test_is_saveable(self, request_mock):
        phone_number = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        phone_number.tags = ["foo", "bar"]
        phone_number.save()
        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_is_modifiable(self, request_mock):
        resource = telnyx.PhoneNumber.modify(TEST_RESOURCE_ID, tags=["foo", "bar"])
        request_mock.assert_requested(
            "patch",
            "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID,
            {"tags": ["foo", "bar"]},
        )
        resource

    @pytest.mark.skip(reason="Prism update")
    def test_is_deletable(self, request_mock):
        resource = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/v2/phone_numbers/%s" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_voice_instance_is_retrievable(self, request_mock):
        phone_number = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        phone_number.voice()

        request_mock.assert_requested(
            "get", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )

    def test_voice_is_listable(self, request_mock):
        resources = telnyx.PhoneNumber.all_voice()

        request_mock.assert_requested("get", "/v2/phone_numbers/voice")

        assert isinstance(resources.data, list)

    @pytest.mark.skip(reason="Prism update")
    def test_voice_instance_is_saveable(self, request_mock):
        phone_number = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        resource = phone_number.voice()
        resource.call_forwarding = {}
        resource.save()

        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_voice_is_saveable(self, request_mock):
        resource = telnyx.PhoneNumber.all_voice().data[0]
        resource.call_forwarding = {}
        resource.save()

        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/voice" % resource.id
        )

    @pytest.mark.skip(reason="Prism update")
    def test_enable_emergency(self, request_mock):
        resource = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        resource = resource.enable_emergency(
            emergency_enabled=True, emergency_address_id="123"
        )
        request_mock.assert_requested(
            "post", "/v2/phone_numbers/%s/actions/enable_emergency" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_messaging_instance_is_retrievable(self, request_mock):
        phone_number = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        phone_number.messaging()

        request_mock.assert_requested(
            "get", "/v2/phone_numbers/%s/messaging" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_messaging_is_listable(self, request_mock):
        resources = telnyx.PhoneNumber.all_messaging()

        request_mock.assert_requested("get", "/v2/phone_numbers/messaging")

        assert isinstance(resources.data, list)

    @pytest.mark.skip(reason="Prism update")
    def test_messaging_instance_is_saveable(self, request_mock):
        phone_number = telnyx.PhoneNumber.retrieve(TEST_RESOURCE_ID)
        resource = phone_number.messaging()
        resource.call_forwarding = {}
        resource.save()

        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/messaging" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Prism update")
    def test_messaging_is_saveable(self, request_mock):
        resource = telnyx.PhoneNumber.all_messaging().data[0]
        resource.call_forwarding = {}
        resource.save()

        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/messaging" % resource.id
        )


class TestVoiceSettings(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.VoiceSettings.list()
        request_mock.assert_requested("get", "/v2/phone_numbers/voice")
        assert isinstance(resources.data, list)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.VoiceSettings.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.VoiceSettings)

    def test_is_saveable(self, request_mock):
        voice_settings = telnyx.VoiceSettings.retrieve(TEST_RESOURCE_ID)
        voice_settings.tech_prefix_enabled = True
        resource = voice_settings.save()
        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.VoiceSettings)

    @pytest.mark.skip(reason="Prism update")
    def test_is_modifiable(self, request_mock):
        resource = telnyx.VoiceSettings.modify(
            TEST_RESOURCE_ID, tech_prefix_enabled=True
        )
        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/voice" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.VoiceSettings)

    def test_instance_url_invalid(self):
        with pytest.raises(error.InvalidRequestError):
            telnyx.VoiceSettings(id=1).instance_url()


class TestMessagingSettings(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.MessagingSettings.list()
        request_mock.assert_requested("get", "/v2/phone_numbers/messaging")
        assert isinstance(resources.data, list)

    def test_is_retrievable(self, request_mock):
        telnyx.MessagingSettings.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/phone_numbers/%s/messaging" % TEST_RESOURCE_ID
        )

    def test_is_saveable(self, request_mock):
        messaging_settings = telnyx.MessagingSettings.retrieve(TEST_RESOURCE_ID)
        messaging_settings.messaging_profile_id = "123"
        resource = messaging_settings.save()
        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/messaging" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.MessagingSettings)

    def test_is_modifiable(self, request_mock):
        resource = telnyx.MessagingSettings.modify(
            TEST_RESOURCE_ID, messaging_profile_id="123"
        )
        request_mock.assert_requested(
            "patch", "/v2/phone_numbers/%s/messaging" % TEST_RESOURCE_ID
        )
        resource

    def test_instance_url_invalid(self):
        with pytest.raises(error.InvalidRequestError):
            telnyx.MessagingSettings(id=1).instance_url()



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import CreateableAPIResource


class PhoneNumberPorting(CreateableAPIResource):
    OBJECT_NAME = "phone_number_porting"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import ListableAPIResource


class WebhookDeliveries(ListableAPIResource):
    OBJECT_NAME = "webhook_deliveries"

    @classmethod
    def class_url(cls):
        return "/v2/webhook_deliveries"



import telnyx

TEST_RESOURCE_ID = "3fa85f64-5717-4562-b3fc-2c963f66afa6"


class TestRoomParticipant(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.RoomParticipant.list()
        request_mock.assert_requested("get", "/v2/room_participants")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.RoomParticipant)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.RoomParticipant.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/room_participants/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.RoomParticipant)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("download", path="download", operations=["list"])
class Media(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "media"

    @classmethod
    def class_url(cls):
        return "/v2/media"

    def download(self, **params):
        return Media.download(self.id, **params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "validate", path="actions/validate", operations=["create"]
)
@nested_resource_class_methods(
    "accept_suggestions", path="actions/accept_suggestions", operations=["create"]
)
class Address(CreateableAPIResource, DeletableAPIResource, ListableAPIResource):
    OBJECT_NAME = "address"

    @classmethod
    def class_url(cls):
        return "/v2/addresses"

    def validate(self, **params):
        return self.create_validate(**params)

    def accept_suggestions(self, **params):
        return self.create_accept_suggestions(**params)



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "6a09cdc3-8948-47f0-aa62-74ac943d6c58"


@pytest.mark.skip(reason="Filter off")
class TestComment(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.Comment.list(
            filter={
                "comment_record_type": "sub_number_order",
                "comment_record_id": "8ffb3622-7c6b-4ccc-b65f-7a3dc0099576",
            }
        )
        request_mock.assert_requested("get", "/v2/comments")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.Comment)

    def test_is_creatable(self, request_mock):
        resource = telnyx.Comment.create()
        request_mock.assert_requested("post", "/v2/comments")
        assert isinstance(resource, telnyx.Comment)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("mno_metadata", path="mnoMetadata", operations=["list"])
@nested_resource_class_methods(
    "operation_status", path="operationStatus", operations=["list"]
)
class Campaign(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "campaign"

    @classmethod
    def class_url(cls):
        return "/10dlc/campaign"

    def mno_metadata(self, **params):
        return Campaign.list_mno_metadata(self.name, **params)

    def operation_status(self, **params):
        return Campaign.list_mno_metadata(self.name, **params)


class CampaignBuilder(CreateableAPIResource):
    OBJECT_NAME = "campaign_builder"


@nested_resource_class_methods("usecase", path="usecase", operations=["retrieve"])
class CampaignBuilderBrand(ListableAPIResource):
    OBJECT_NAME = "campaign_builder_brand"

    def download(self, **params):
        return Campaign.create_usecase(self.id, **params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "refresh_client_token", path="actions/refresh_client_token", operations=["create"]
)
@nested_resource_class_methods(
    "generate_join_client_token",
    path="actions/generate_join_client_token",
    operations=["create"],
)
class RoomsClientToken(CreateableAPIResource):
    OBJECT_NAME = "rooms_client_token"

    def refresh_client_token(self, **params):
        return self.create_refresh_client_token(**params)

    def generate_join_client_token(self, **params):
        return self.create_generate_join_client_token(**params)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    ListableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("join", path="actions/join", operations=["create"])
@nested_resource_class_methods("mute", path="actions/mute", operations=["create"])
@nested_resource_class_methods("unmute", path="actions/unmute", operations=["create"])
@nested_resource_class_methods("hold", path="actions/hold", operations=["create"])
@nested_resource_class_methods("unhold", path="actions/unhold", operations=["create"])
@nested_resource_class_methods("speak", path="actions/speak", operations=["create"])
@nested_resource_class_methods("play", path="actions/play", operations=["create"])
@nested_resource_class_methods("stop", path="actions/stop", operations=["create"])
@nested_resource_class_methods(
    "record_start", path="actions/record_start", operations=["create"]
)
@nested_resource_class_methods(
    "record_stop", path="actions/record_stop", operations=["create"]
)
@nested_resource_class_methods("update", path="actions/update", operations=["create"])
@nested_resource_class_methods("leave", path="actions/leave", operations=["create"])
@nested_resource_class_methods(
    "dial_participant", path="actions/dial_participant", operations=["create"]
)
@nested_resource_class_methods("participants", path="participants", operations=["list"])
class Conference(CreateableAPIResource, ListableAPIResource):
    OBJECT_NAME = "conference"

    def join(self, **params):
        return Conference.create_join(self.id, **params)

    def dial_participant(self, **params):
        return Conference.create_dial_participant(self.id, **params)

    def mute(self, **params):
        return Conference.create_mute(self.id, **params)

    def unmute(self, **params):
        return Conference.create_unmute(self.id, **params)

    def hold(self, **params):
        return Conference.create_hold(self.id, **params)

    def unhold(self, **params):
        return Conference.create_unhold(self.id, **params)

    def speak(self, **params):
        return Conference.create_speak(self.id, **params)

    def play(self, **params):
        return Conference.create_play(self.call_control_id, **params)

    def stop(self, **params):
        return Conference.create_stop(self.call_control_id, **params)

    def record_start(self, **params):
        return Conference.create_record_start(self.id, **params)

    def record_stop(self, **params):
        return Conference.create_record_stop(self.id, **params)

    def update(self, **params):
        return Conference.create_update(self.id, **params)

    def leave(self, **params):
        return Conference.create_leave(self.id, **params)

    def participants(self, **params):
        return Conference.create_participants(self.name, **params)



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

NUMBER_RESERVATION_ID = "12ade33a-21c0-473b-b055-b3c836e1c292"


def create_number_reservation():
    return telnyx.NumberReservation.create(
        phone_numbers=[{"phone_number": "++19705555098"}]
    )


class TestNumberReservation(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.NumberReservation.list()
        request_mock.assert_requested("get", "/v2/number_reservations")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.NumberReservation)

    def test_is_retrievable(self, request_mock):
        resources = telnyx.NumberReservation.retrieve(NUMBER_RESERVATION_ID)
        request_mock.assert_requested(
            "get", "/v2/number_reservations/{}".format(NUMBER_RESERVATION_ID)
        )
        assert isinstance(resources, telnyx.NumberReservation)

    def test_is_creatable(self, request_mock):
        resource = create_number_reservation()
        request_mock.assert_requested("post", "/v2/number_reservations")
        assert isinstance(resource, telnyx.NumberReservation)

    @pytest.mark.skip(reason="async, wait for response")
    def test_can_call_extend(self, request_mock):
        resource = create_number_reservation()
        resource.extend()
        request_mock.assert_requested(
            "post", "/v2/number_reservations/%s/actions/extend" % NUMBER_RESERVATION_ID
        )
        assert isinstance(resource, telnyx.NumberReservation)

    @pytest.mark.skip(reason="async, wait for response")
    def test_can_call_reservation_extend(self, request_mock):
        resource = create_number_reservation()
        resource.create_extend(NUMBER_RESERVATION_ID)
        request_mock.assert_requested(
            "post", "/v2/number_reservations/%s/actions/extend" % NUMBER_RESERVATION_ID
        )
        assert isinstance(resource, telnyx.NumberReservation)



from __future__ import absolute_import, division, print_function

import telnyx


class TestNumberLookup(object):
    def test_is_retrievable(self, request_mock):
        resource = telnyx.NumberLookup.retrieve("+18665552368")
        request_mock.assert_requested("get", "/v2/number_lookup/%2B18665552368")
        assert isinstance(resource, telnyx.NumberLookup)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("verify", path="actions/verify", operations=["create"])
class VerifiedNumber(CreateableAPIResource, DeletableAPIResource, ListableAPIResource):
    OBJECT_NAME = "verified_number"

    def verify(self, **params):
        return self.create_verify(**params)



from __future__ import absolute_import, division, print_function

from telnyx import api_requestor, util
from telnyx.api_resources.abstract.api_resource import APIResource


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, api_key=None, **params):
        requestor = api_requestor.APIRequestor(api_key)
        url = cls.class_url()
        params = util.rewrite_reserved_words(params)
        response, api_key = requestor.request("post", url, params)

        return util.convert_to_telnyx_object(response, api_key)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    APIResource,
    ListableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods("mute", path="actions/mute", operations=["create"])
@nested_resource_class_methods("unmute", path="actions/unmute", operations=["create"])
@nested_resource_class_methods("kick", path="actions/kick", operations=["create"])
@nested_resource_class_methods(
    "participants", path="participants", operations=["list", "retrieve"]
)
class RoomSession(ListableAPIResource):
    OBJECT_NAME = "room_session"

    def mute(self, **params):
        return RoomSession.create_mute(self.id, **params)

    def unmute(self, **params):
        return RoomSession.create_unmute(self.id, **params)

    def kick(self, **params):
        return RoomSession.create_kick(self.id, **params)

    def list_participants(self, **params):
        return RoomSession.list_participants(self.name, **params)


class RoomSessionList(APIResource):
    OBJECT_NAME = "room_session_list"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.access_control_ip import AccessControlIP
from telnyx.api_resources.access_ip_address import AccessIPAddress
from telnyx.api_resources.access_token import AccessToken
from telnyx.api_resources.address import Address
from telnyx.api_resources.advanced_optinoptout import AdvancedOptinoptout
from telnyx.api_resources.ai import AI
from telnyx.api_resources.api_key import APIKey
from telnyx.api_resources.authentication_provider import AuthenticationProvider
from telnyx.api_resources.autorechargepreference import Autorechargepreference
from telnyx.api_resources.available_phone_number import AvailablePhoneNumber
from telnyx.api_resources.balance import Balance
from telnyx.api_resources.billing import Billing
from telnyx.api_resources.billing_group import BillingGroup
from telnyx.api_resources.brand import Brand
from telnyx.api_resources.bucket import Bucket
from telnyx.api_resources.bucket_usage import BucketUsage
from telnyx.api_resources.bulk_creation import BulkCreation
from telnyx.api_resources.bulk_credential import BulkCredential
from telnyx.api_resources.bulk_phone_number_campaign import BulkPhoneNumberCampaign
from telnyx.api_resources.bulk_phone_number_operation import BulkPhoneNumberOperation
from telnyx.api_resources.bulk_sole_proprietor_creation import (
    BulkSoleProprietorCreation,
)
from telnyx.api_resources.business_identity import BusinessIdentity
from telnyx.api_resources.call import Call
from telnyx.api_resources.call_control_application import CallControlApplication
from telnyx.api_resources.call_information import CallInformation
from telnyx.api_resources.call_recording import CallRecording
from telnyx.api_resources.campaign import Campaign
from telnyx.api_resources.cdr_usage_report import CdrUsageReport
from telnyx.api_resources.comment import Comment
from telnyx.api_resources.conference import Conference
from telnyx.api_resources.connection import Connection
from telnyx.api_resources.credential_connection import CredentialConnection
from telnyx.api_resources.csv_download import CsvDownload
from telnyx.api_resources.custom_storage_credential import CustomStorageCredential
from telnyx.api_resources.customer_service_record import CustomerServiceRecord
from telnyx.api_resources.debugging import Debugging
from telnyx.api_resources.detail_record import DetailRecord
from telnyx.api_resources.detail_records_report import DetailRecordsReport
from telnyx.api_resources.dialogflow_integration import DialogflowIntegration
from telnyx.api_resources.document import Document, DocumentLink
from telnyx.api_resources.dynamic_emergency_address import DynamicEmergencyAddress
from telnyx.api_resources.dynamic_emergency_endpoint import DynamicEmergencyEndpoint
from telnyx.api_resources.enum import Enum
from telnyx.api_resources.event import Event
from telnyx.api_resources.external_connection import ExternalConnection
from telnyx.api_resources.fax import Fax
from telnyx.api_resources.fax_application import FaxApplication
from telnyx.api_resources.fqdn import FQDN
from telnyx.api_resources.fqdn_connection import FQDNConnection
from telnyx.api_resources.inbound_channel import InboundChannel
from telnyx.api_resources.inventory_coverage import InventoryCoverage
from telnyx.api_resources.inventory_level import InventoryLevel
from telnyx.api_resources.ip import IP
from telnyx.api_resources.ip_address import IPAddress
from telnyx.api_resources.ip_connection import IPConnection
from telnyx.api_resources.ip_range import IPRange
from telnyx.api_resources.list_object import ListObject
from telnyx.api_resources.managed_account import ManagedAccount
from telnyx.api_resources.mdr_detail_report import MdrDetailReport
from telnyx.api_resources.mdr_usage_report import MdrUsageReport
from telnyx.api_resources.media import Media
from telnyx.api_resources.media_storage_api import MediaStorageApi
from telnyx.api_resources.message import Message
from telnyx.api_resources.messaging_hosted_number import MessagingHostedNumber
from telnyx.api_resources.messaging_hosted_number_order import (
    MessagingHostedNumberOrder,
)
from telnyx.api_resources.messaging_phone_number import MessagingPhoneNumber
from telnyx.api_resources.messaging_profile import MessagingProfile
from telnyx.api_resources.messaging_tollfree_verification import (
    MessagingTollfreeVerification,
)
from telnyx.api_resources.messaging_url_domain import MessagingUrlDomain
from telnyx.api_resources.mobile_network_operator import MobileNetworkOperator
from telnyx.api_resources.network import Network
from telnyx.api_resources.notification import Notification
from telnyx.api_resources.notification_channel import NotificationChannel
from telnyx.api_resources.notification_event import NotificationEvent
from telnyx.api_resources.notification_event_condition import NotificationEventCondition
from telnyx.api_resources.notification_profiles import NotificationProfile
from telnyx.api_resources.notification_setting import NotificationSetting
from telnyx.api_resources.number_configuration import NumberConfiguration
from telnyx.api_resources.number_lookup import NumberLookup
from telnyx.api_resources.number_order import NumberOrder
from telnyx.api_resources.number_order_phone_number import NumberOrderPhoneNumber
from telnyx.api_resources.number_portout import NumberPortout
from telnyx.api_resources.number_reservation import NumberReservation
from telnyx.api_resources.numbers_feature import NumbersFeature
from telnyx.api_resources.ota_update import OtaUpdate
from telnyx.api_resources.outbound_voice_profile import OutboundVoiceProfile
from telnyx.api_resources.phone_assignment_profile import PhoneNumberByProfile
from telnyx.api_resources.phone_number import (
    MessagingSettings,
    PhoneNumber,
    VoiceSettings,
)
from telnyx.api_resources.phone_number_block_order import PhoneNumberBlockOrder
from telnyx.api_resources.phone_number_blocks_background_job import (
    PhoneNumberBlocksBackgroundJob,
)
from telnyx.api_resources.phone_number_campaign import PhoneNumberCampaign
from telnyx.api_resources.phone_number_configuration import PhoneNumberConfiguration
from telnyx.api_resources.phone_number_jobs import PhoneNumberJob
from telnyx.api_resources.phone_number_order import PhoneNumberOrder
from telnyx.api_resources.phone_number_order_document import PhoneNumberOrderDocument
from telnyx.api_resources.phone_number_porting import PhoneNumberPorting
from telnyx.api_resources.phone_number_regulatory_requirement import (
    PhoneNumberRegulatoryRequirement,
)
from telnyx.api_resources.phone_number_reservation import PhoneNumberReservation
from telnyx.api_resources.phone_number_search import PhoneNumberSearch
from telnyx.api_resources.portability_check import PortabilityCheck
from telnyx.api_resources.porting_order import PortingOrder, PortingPhoneNumber
from telnyx.api_resources.portout import PortOut
from telnyx.api_resources.presigned_object_url import PresignedObjectUrl
from telnyx.api_resources.private_wireless_gateway import PrivateWirelessGateway
from telnyx.api_resources.public_internet_gateway import PublicInternetGateway
from telnyx.api_resources.push_credential import PushCredential
from telnyx.api_resources.queue import Queue, QueueCall
from telnyx.api_resources.queue_command import QueueCommand
from telnyx.api_resources.recordings_command import RecordingsCommand
from telnyx.api_resources.region import Region
from telnyx.api_resources.register_call import RegisterCall
from telnyx.api_resources.report import Report
from telnyx.api_resources.reporting import Reporting
from telnyx.api_resources.requirement import Requirement
from telnyx.api_resources.requirement_type import RequirementType
from telnyx.api_resources.room import Room
from telnyx.api_resources.room_composition import RoomComposition
from telnyx.api_resources.room_participant import RoomParticipant
from telnyx.api_resources.room_recording import RoomRecording
from telnyx.api_resources.room_session import RoomSession
from telnyx.api_resources.rooms_client_token import RoomsClientToken
from telnyx.api_resources.shared_campaign import SharedCampaign
from telnyx.api_resources.short_code import ShortCode
from telnyx.api_resources.sim_card import SIMCard
from telnyx.api_resources.sim_card_action import SIMCardAction
from telnyx.api_resources.sim_card_group import SIMCardGroup
from telnyx.api_resources.sim_card_order import SIMCardOrder, SIMCardOrderPreview
from telnyx.api_resources.ssl_certificate import SslCertificate
from telnyx.api_resources.sub_number_order import SubNumberOrder
from telnyx.api_resources.telephony_credential import TelephonyCredential
from telnyx.api_resources.texml_application import TexmlApplication
from telnyx.api_resources.texml_rest_command import TexmlRestCommand
from telnyx.api_resources.verification import Verification
from telnyx.api_resources.verified_calls_display_profile import (
    VerifiedCallsDisplayProfile,
)
from telnyx.api_resources.verified_number import VerifiedNumber
from telnyx.api_resources.verify import Verify
from telnyx.api_resources.verify_profile import VerifyProfile
from telnyx.api_resources.virtual_cross_connect import VirtualCrossConnect
from telnyx.api_resources.wdr_detail_report import WdrDetailReport
from telnyx.api_resources.webhook import Webhook
from telnyx.api_resources.webhook_deliveries import WebhookDeliveries
from telnyx.api_resources.wireguard_interface import WireguardInterface
from telnyx.api_resources.wireless_detail_record_report import (
    WirelessDetailRecordsReports,
)

# flake8: noqa


__all__ = [
    "APIKey",
    "AccessControlIP",
    "AccessIPAddress",
    "AccessToken",
    "Address",
    "AdvancedOptinoptout",
    "AI",
    "AuthenticationProvider",
    "Autorechargepreference",
    "AvailablePhoneNumber",
    "Balance",
    "Billing",
    "BillingGroup",
    "Brand",
    "Bucket",
    "BucketUsage",
    "BulkCreation",
    "BulkCredential",
    "BulkPhoneNumberCampaign",
    "BulkPhoneNumberOperation",
    "BulkSoleProprietorCreation",
    "BusinessIdentity",
    "Call",
    "CallControlApplication",
    "CallInformation",
    "CallRecording",
    "Campaign",
    "CdrUsageReport",
    "Comment",
    "Conference",
    "Connection",
    "CredentialConnection",
    "CsvDownload",
    "CustomStorageCredential",
    "CustomerServiceRecord",
    "Debugging",
    "DetailRecord",
    "DetailRecordsReport",
    "DialogflowIntegration",
    "Document",
    "DocumentLink",
    "DynamicEmergencyAddress",
    "DynamicEmergencyEndpoint",
    "Enum",
    "Event",
    "ExternalConnection",
    "Fax",
    "FaxApplication",
    "FQDN",
    "FQDNConnection",
    "IP",
    "IPAddress",
    "IPConnection",
    "IPRange",
    "InboundChannel",
    "InventoryCoverage",
    "InventoryLevel",
    "ListObject",
    "ManagedAccount",
    "MdrDetailReport",
    "MdrUsageReport",
    "Media",
    "MediaStorageApi",
    "Message",
    "MessagingHostedNumber",
    "MessagingHostedNumberOrder",
    "MessagingPhoneNumber",
    "MessagingProfile",
    "MessagingSettings",
    "MessagingTollfreeVerification",
    "MessagingUrlDomain",
    "MobileNetworkOperator",
    "Network",
    "Notification",
    "NotificationChannel",
    "NotificationEvent",
    "NotificationEventCondition",
    "NotificationProfile",
    "NotificationSetting",
    "NumberConfiguration",
    "NumberLookup",
    "NumberOrder",
    "NumberOrderPhoneNumber",
    "NumberPortout",
    "NumberReservation",
    "NumbersFeature",
    "OtaUpdate",
    "OutboundVoiceProfile",
    "PhoneNumber",
    "PhoneNumberBlockOrder",
    "PhoneNumberBlocksBackgroundJob",
    "PhoneNumberByProfile",
    "PhoneNumberCampaign",
    "PhoneNumberConfiguration",
    "PhoneNumberJob",
    "PhoneNumberOrder",
    "PhoneNumberOrderDocument",
    "PhoneNumberPorting",
    "PhoneNumberRegulatoryRequirement",
    "PhoneNumberReservation",
    "PhoneNumberSearch",
    "PortOut",
    "PortabilityCheck",
    "PortingOrder",
    "PortingPhoneNumber",
    "PresignedObjectUrl",
    "PrivateWirelessGateway",
    "PublicInternetGateway",
    "PushCredential",
    "Queue",
    "QueueCall",
    "QueueCommand",
    "RecordingsCommand",
    "Region",
    "RegisterCall",
    "Report",
    "Reporting",
    "Requirement",
    "RequirementType",
    "Room",
    "RoomComposition",
    "RoomParticipant",
    "RoomRecording",
    "RoomSession",
    "RoomsClientToken",
    "SIMCard",
    "SIMCardAction",
    "SIMCardGroup",
    "SIMCardOrder",
    "SIMCardOrderPreview",
    "SharedCampaign",
    "ShortCode",
    "SslCertificate",
    "SubNumberOrder",
    "TelephonyCredential",
    "TexmlApplication",
    "TexmlRestCommand",
    "Verification",
    "VerifiedCallsDisplayProfile",
    "VerifiedNumber",
    "Verify",
    "VerifyProfile",
    "VirtualCrossConnect",
    "VoiceSettings",
    "WdrDetailReport",
    "Webhook",
    "WebhookDeliveries",
    "WireguardInterface",
    "WirelessDetailRecordsReports",
]



from __future__ import absolute_import, division, print_function

import pytest
import telnyx

TEST_RESOURCE_ID = "1293384261075731499"


class TestAddress(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.Address.list()
        request_mock.assert_requested("get", "/v2/addresses")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.Address)

    @pytest.mark.skip(reason="Prism mock 500 invalid response")
    def test_is_retrievable(self, request_mock):
        resource = telnyx.Address.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/addresses/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.Address)

    def test_is_creatable(self, request_mock):
        resource = telnyx.Address.create(
            name="my-profile",
            business_name="Acme Inc",
            country_code="US",
            street_address="311 W Superior St",
            first_name="John",
            last_name="Doe",
            locality="Chicago",
        )
        request_mock.assert_requested("post", "/v2/addresses")
        assert isinstance(resource, telnyx.Address)

    def test_is_deletable(self, request_mock):
        resource = telnyx.Address(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested("delete", "/v2/addresses/%s" % TEST_RESOURCE_ID)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    nested_resource_class_methods,
)


@nested_resource_class_methods(
    "remove_private_wireless_gateway",
    path="actions/remove_private_wireless_gateway",
    operations=["create"],
)
@nested_resource_class_methods(
    "set_private_wireless_gateway",
    path="actions/set_private_wireless_gateway",
    operations=["create"],
)
class SIMCardGroup(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "sim_card_group"

    def remove_private_wireless_gateway(self, **params):
        return self.create_remove_private_wireless_gateway(**params)

    def set_private_wireless_gateway(self, **params):
        return self.create_set_private_wireless_gateway(**params)



from __future__ import absolute_import, division, print_function

import telnyx

TEST_RESOURCE_ID = "f5586561-8ff0-4291-a0ac-84fe544797bd"


class TestBillingGroup(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.BillingGroup.list()
        request_mock.assert_requested("get", "/v2/billing_groups")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.BillingGroup)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.BillingGroup.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/billing_groups/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.BillingGroup)

    def test_is_creatable(self, request_mock):
        resource = telnyx.BillingGroup.create(name="foo")
        request_mock.assert_requested("post", "/v2/billing_groups")
        assert isinstance(resource, telnyx.BillingGroup)

    def test_is_saveable(self, request_mock):
        billing_group = telnyx.BillingGroup.retrieve(TEST_RESOURCE_ID)
        billing_group.name = "bar"
        resource = billing_group.save()
        request_mock.assert_requested(
            "patch", "/v2/billing_groups/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.BillingGroup)
        assert resource is billing_group

    def test_is_modifiable(self, request_mock):
        resource = telnyx.BillingGroup.modify(TEST_RESOURCE_ID, name="foo")
        request_mock.assert_requested(
            "patch", "/v2/billing_groups/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.BillingGroup)

    def test_is_deletable(self, request_mock):
        resource = telnyx.BillingGroup.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/v2/billing_groups/%s" % TEST_RESOURCE_ID
        )



from __future__ import absolute_import, division, print_function

from six import python_2_unicode_compatible


@python_2_unicode_compatible
class TelnyxError(Exception):
    def __init__(
        self,
        errors=None,
        http_status=None,
        http_body=None,
        json_body=None,
        http_headers=None,
    ):
        self.errors = errors
        self.http_status = http_status
        self.json_body = json_body
        self.http_headers = http_headers or {}
        self.request_id = self.http_headers.get("request-id", None)
        message = self.build_message()
        self._message = message

        super(TelnyxError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as UTF-8. "
                    "Please report to support@telnyx.com>"
                )
        self.http_body = http_body

    def build_message(self):
        return "{message}{other_errors_message}{full_details}".format(
            message=self.extract_message(),
            other_errors_message=self.other_errors_message(),
            full_details=self.full_details(),
        )

    def full_details(self):
        if isinstance(self.errors, list) and len(self.errors) > 0:
            return "Full details: {errors}".format(errors=self.errors)
        else:
            return ""

    def extract_message(self):
        if isinstance(self.errors, list):
            if len(self.errors) > 0:
                return self.errors[0].get("title", "")
            else:
                return ""
        elif self.errors is None:
            return ""
        else:
            return self.errors

    def other_errors_message(self):
        count = self.errors_count()
        if count > 2:
            return "plus {count} other errors. ".format(count=count)
        elif count == 2:
            return "plus 1 other error. "
        else:
            return ""

    def errors_count(self):
        if isinstance(self.errors, list):
            return len(self.errors)
        else:
            return 1

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return "Request {0}: {1}".format(self.request_id, msg)
        else:
            return msg

    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by Telnyx's API. This was previously available
    # in python2 via `error.message`. Unlike `str(error)`, it omits "Request
    # req_..." from the beginning of the string.
    @property
    def user_message(self):
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )


class APIConnectionError(TelnyxError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        http_headers=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, http_headers
        )
        self.should_retry = should_retry


class InvalidRequestError(TelnyxError):
    # 400
    pass


class AuthenticationError(TelnyxError):
    # 401
    pass


class PermissionError(TelnyxError):
    # 403
    pass


class ResourceNotFoundError(TelnyxError):
    # 404
    pass


class MethodNotSupportedError(TelnyxError):
    # 405
    pass


class TimeoutError(TelnyxError):
    # 408
    pass


class UnsupportedMediaTypeError(TelnyxError):
    # 415
    pass


class InvalidParametersError(TelnyxError):
    # 422
    pass


class RateLimitError(TelnyxError):
    # 429
    pass


class APIError(TelnyxError):
    # 500
    pass


class ServiceUnavailableError(TelnyxError):
    # 503
    pass


class SignatureVerificationError(TelnyxError):
    def __init__(self, message, sig_header, timestamp_header, http_body=None):
        super(SignatureVerificationError, self).__init__(message, http_body)
        self.sig_header = sig_header
        self.timestamp_header = timestamp_header



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "45f45a04-b4be-4592-95b1-9306b9db2b21"


@pytest.mark.skip(reason="Filter looking for 'filter%5bphone_number%5d', typo")
class TestInventoryCoverage(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.InventoryCoverage.list(filter={"groupBy": "npa"})
        request_mock.assert_requested("get", "/v2/inventory_coverage")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.InventoryCoverage)



import pytest

import telnyx

TEST_RESOURCE_ID = "3fa85f64-5717-4562-b3fc-2c963f66afa6"


class TestRoomSessions(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.RoomSession.list()
        request_mock.assert_requested("get", "/v2/room_sessions")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.RoomSession)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.RoomSession.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/room_sessions/%s" % TEST_RESOURCE_ID)
        assert isinstance(resource, telnyx.RoomSession)

    def test_can_call_create_mute(self, request_mock):
        resource = telnyx.RoomSession.retrieve(TEST_RESOURCE_ID)
        resource.create_mute(id=TEST_RESOURCE_ID, particpants="all")
        request_mock.assert_requested(
            "post", "/v2/room_sessions/%s/actions/mute" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.RoomSession)

    def test_can_call_create_unmute(self, request_mock):
        resource = telnyx.RoomSession.retrieve(TEST_RESOURCE_ID)
        resource.create_unmute(id=TEST_RESOURCE_ID, particpants="all")
        request_mock.assert_requested(
            "post", "/v2/room_sessions/%s/actions/unmute" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.RoomSession)

    def test_can_call_kick(self, request_mock):
        resource = telnyx.RoomSession.retrieve(TEST_RESOURCE_ID)
        resource.create_kick(id=TEST_RESOURCE_ID, particpants="all")
        request_mock.assert_requested(
            "post", "/v2/room_sessions/%s/actions/kick" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.RoomSession)

    @pytest.mark.skip(reason="Participant list ")
    def test_can_list_participants(self, request_mock):
        resource = telnyx.RoomSession.retrieve(TEST_RESOURCE_ID)
        resource.active = True
        participants = resource.list_participants()
        request_mock.assert_requested(
            "get", "/v2/room_sessions/%s/participants" % resource.name
        )
        assert isinstance(participants.data[0], telnyx.RoomSession)



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import ListableAPIResource, UpdateableAPIResource


class ShortCode(ListableAPIResource, UpdateableAPIResource):
    OBJECT_NAME = "short_code"



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import (
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
)


class WirelessDetailRecordsReports(
    CreateableAPIResource, ListableAPIResource, DeletableAPIResource
):
    OBJECT_NAME = "wireless_detail_records_report"

    @classmethod
    def class_url(cls):
        return "/v2/wireless/detail_records_reports"



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "1d12cb27-4cde-8e18-0b7c-fc1e43685fde"


class TestPortingOrder(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.PortingOrder.list()
        request_mock.assert_requested("get", "/v2/porting_orders")
        assert isinstance(resources.data, list)

    def test_is_retrievable(self, request_mock):
        telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested("get", "/v2/porting_orders/%s" % TEST_RESOURCE_ID)

    def test_is_creatable(self, request_mock):
        resource = telnyx.PortingOrder.create(
            phone_numbers=["13035550000", "13035550001", "13035550002"]
        )
        resource
        request_mock.assert_requested("post", "/v2/porting_orders")

    @pytest.mark.skip(reason="Needs live confirm")
    def test_is_saveable(self, request_mock):
        porting_order = telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        porting_order.webhook_event = "https://update.com"
        porting_order.customer_reference = "updated name"
        porting_order.save()
        request_mock.assert_requested(
            "patch", "/v2/porting_orders/%s" % TEST_RESOURCE_ID
        )

    def test_is_modifiable(self, request_mock):
        resource = telnyx.PortingOrder.modify(
            TEST_RESOURCE_ID,
            webhook_event="https://update.com",
            customer_reference="updated name",
        )
        resource
        request_mock.assert_requested(
            "patch", "/v2/porting_orders/%s" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Needs live confirm")
    def test_can_confirm_porting_order_action(self, request_mock):
        resource = telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        resource.create_confirm(id=TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v2/porting_orders/%s/actions/confirm" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Needs live confirm")
    def test_can_cancel_porting_order_action(self, request_mock):
        resource = telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        resource.create_cancel(id=TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v2/porting_orders/%s/actions/cancel" % TEST_RESOURCE_ID
        )

    @pytest.mark.skip(reason="Needs live confirm")
    def test_can_get_loa_template(self, request_mock):
        resource = telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        resource.loaTemplate()
        request_mock.assert_requested(
            "get", "/v2/porting_orders/%s/loa_template" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.PortingOrder)

    @pytest.mark.skip(reason="Needs live confirm")
    def test_can_list_allowed_foc_dates(self, request_mock):
        resource = telnyx.PortingOrder.retrieve(TEST_RESOURCE_ID)
        resource.allowed_foc_windows()
        request_mock.assert_requested(
            "get", "/v2/porting_orders/%s/allowed_foc_windows" % TEST_RESOURCE_ID
        )

    def test_can_list_porting_phone_numbers(self, request_mock):
        resource = telnyx.PortingPhoneNumber.list()
        request_mock.assert_requested("get", "/v2/porting_phone_numbers")
        assert isinstance(resource.data, list)



from __future__ import absolute_import, division, print_function

import telnyx

TEST_RESOURCE_ID = "6a09cdc3-8948-47f0-aa62-74ac943d6c58"


class TestSIMCardActions(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.SIMCardAction.list()
        request_mock.assert_requested("get", "/v2/sim_card_actions")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.SIMCardAction)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.SIMCardAction.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v2/sim_card_actions/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.SIMCardAction)



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import quote_plus

from telnyx import util
from telnyx.telnyx_object import TelnyxObject


class ListObject(TelnyxObject):
    def list(self, **params):
        r = self.request("get", self["url"], params)
        r["url"] = self["url"]

        return r

    @classmethod
    def empty_list(cls, params, url):
        list_object = ListObject()
        list_object.retrieve_params = params
        list_object.url = url
        list_object.data = []
        return list_object

    def auto_paging_iter(self):
        page = self
        params = dict(self._retrieve_params)

        while True:
            for item in page.data:
                yield item
            if page.empty():
                return

            page = page.next_page(**params)

    def auto_paging_iter_by_token(self):
        page = self
        params = dict(self._retrieve_params)

        while True:
            for item in page.data:
                yield item
            if page.empty():
                return

            page = page.next_page_by_token(**params)

    def empty(self):
        return len(self.data) == 0

    def next_page(self, **params):
        if not self.has_more():
            return ListObject.empty_list(params, self.url)

        next_page_number = self.page_number() + 1
        pagination = {"number": next_page_number, "size": self.page_size()}
        params["page"] = pagination

        return self.list(**params)

    def next_page_by_token(self, **params):
        if self.token() is None:
            return ListObject.empty_list(params, self.url)

        pagination = {"token": self.token()}
        params["page"] = pagination

        return self.list(**params)

    def previous_page(self, **params):
        prev_page_number = self.page_number() - 1
        prev_page_number = max(prev_page_number, 1)
        pagination = {"number": prev_page_number, "size": self.page_size()}
        params["page"] = pagination

        return self.list(**params)

    def has_more(self):
        return self.data != [] and self.get("meta", {}).get(
            "total_pages", 0
        ) > self.get("meta", {}).get("page_number", 0)

    def token(self):
        return self.get("meta", {}).get("next_page_token", None)

    def page_number(self):
        return self.get("meta", {}).get("page_number", 1)

    def page_size(self):
        return self.get("meta", {}).get("page_size", 20)

    def create(self, **params):
        return self.request("post", self["url"], params)

    def retrieve(self, id, **params):
        base = self.get("url")
        id = util.utf8(id)
        extn = quote_plus(id)
        url = "%s/%s" % (base, extn)

        return self.request("get", url, params)

    def __iter__(self):
        return getattr(self, "data", []).__iter__()

    def __len__(self):
        return getattr(self, "data", []).__len__()



from __future__ import absolute_import, division, print_function

from telnyx.api_resources.abstract import ListableAPIResource


class QueueCommand(ListableAPIResource):
    OBJECT_NAME = "queue_command"



from __future__ import absolute_import, division, print_function

import pytest

import telnyx

TEST_RESOURCE_ID = "1293384261075731499"


@pytest.mark.skip(reason="Unreleased")
class TestPhoneNumberCampaign(object):
    def test_is_listable(self, request_mock):
        resources = telnyx.PhoneNumberCampaign.list()
        request_mock.assert_requested("get", "/10dlc/phoneNumberCampaign")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], telnyx.PhoneNumberCampaign)

    def test_is_retrievable(self, request_mock):
        resource = telnyx.PhoneNumberCampaign.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/10dlc/phoneNumberCampaign/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.PhoneNumberCampaign)

    def test_is_creatable(self, request_mock):
        resource = telnyx.PhoneNumberCampaign.create(
            phoneNumber="1920324562", campaignID=TEST_RESOURCE_ID
        )
        request_mock.assert_requested("post", "/10dlc/phoneNumberCampaign")
        assert isinstance(resource, telnyx.PhoneNumberCampaign)

    def test_is_modifiable(self, request_mock):
        resource = telnyx.PhoneNumberCampaign.modify(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "patch", "/10dlc/phoneNumberCampaign/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, telnyx.PhoneNumberCampaign)

    def test_is_deletable(self, request_mock):
        resource = telnyx.PhoneNumberCampaign.retrieve(TEST_RESOURCE_ID)
        resource.delete()
        request_mock.assert_requested(
            "delete", "/10dlc/phoneNumberCampaign/%s" % TEST_RESOURCE_ID
        )



from __future__ import absolute_import, division, print_function

import json

from telnyx.telnyx_response import TelnyxResponse


class TestTelnyxResponse(object):
    def test_request_id(self):
        response, headers, body, code = self.mock_telnyx_response()
        assert response.request_id == headers["request-id"]

        del headers["request-id"]

        assert response.request_id is None

    def test_code(self):
        response, headers, body, code = self.mock_telnyx_response()
        assert response.code == code

    def test_headers(self):
        response, headers, body, code = self.mock_telnyx_response()
        assert response.headers == headers

    def test_body(self):
        response, headers, body, code = self.mock_telnyx_response()
        assert response.body == body

    def test_data(self):
        response, headers, body, code = self.mock_telnyx_response()
        assert response.data == json.loads(body)

    @staticmethod
    def mock_telnyx_response():
        code = 200
        headers = TestTelnyxResponse.mock_headers()
        body = TestTelnyxResponse.mock_body()
        response = TelnyxResponse(body, code, headers)
        return response, headers, body, code

    @staticmethod
    def mock_headers():
        return {"request-id": "123456"}

    @staticmethod
    def mock_body():
        return '{"data": { "id": "12345", "record_type": "messaging_profile", "name": "New Messaging Profile" }}'



from __future__ import absolute_import, division, print_function

import calendar
import datetime
import json
import platform
import time

import six
from six.moves.urllib.parse import urlencode, urlsplit, urlunsplit

import telnyx
from telnyx import error, http_client, util
from telnyx.multipart_data_generator import MultipartDataGenerator
from telnyx.telnyx_response import TelnyxResponse


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = {}
    for subkey, subvalue in six.iteritems(data):
        d[fmt % (key, subkey)] = subvalue
    return d


def _api_encode(data):
    for key, value in six.iteritems(data):
        key = util.utf8(key)
        if value is None:
            continue
        elif isinstance(value, list) or isinstance(value, tuple):
            for sv in value:
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[]" % key, sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[]" % key, util.utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, util.utf8(value))


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):
    def __init__(self, key=None, client=None, api_base=None):
        self.api_base = api_base or telnyx.api_base
        self.api_key = key

        from telnyx import proxy, verify_ssl_certs as verify

        self._client = (
            client
            or telnyx.default_http_client
            or http_client.new_default_http_client(verify_ssl_certs=verify, proxy=proxy)
        )

        self._last_request_metrics = None

    @classmethod
    def format_app_info(cls, info):
        str = info["name"]
        if info["version"]:
            str += "/%s" % (info["version"],)
        if info["url"]:
            str += " (%s)" % (info["url"],)
        return str

    def request(self, method, url, params=None, headers=None):
        rbody, rcode, rheaders, my_api_key = self.request_raw(
            method.lower(), url, params, headers
        )
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp, my_api_key

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            error_list = resp["errors"]
        except (KeyError, TypeError):
            raise error.APIError(
                [
                    {
                        "title": "Invalid response object from API: %r (HTTP response code "
                        "was %d)" % (rbody, rcode)
                    }
                ],
                rbody,
                rcode,
                resp,
            )

        err = self.specific_api_error(rbody, rcode, resp, rheaders, error_list)

        raise err

    def specific_api_error(self, rbody, rcode, resp, rheaders, error_list):
        for err in error_list:
            util.log_info(
                "Telnyx API error received",
                error_code=err.get("code"),
                error_detail=err.get("detail"),
                error_source=err.get("source"),
                error_title=err.get("title"),
            )

        if rcode == 400:
            return error.InvalidRequestError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 401:
            return error.AuthenticationError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 403:
            return error.PermissionError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 404:
            return error.ResourceNotFoundError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 405:
            return error.MethodNotSupportedError(
                error_list, rcode, rbody, resp, rheaders
            )
        elif rcode == 408:
            return error.TimeoutError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 415:
            return error.UnsupportedMediaTypeError(
                error_list, rcode, rbody, resp, rheaders
            )
        elif rcode == 422:
            return error.InvalidParametersError(
                error_list, rcode, rbody, resp, rheaders
            )
        elif rcode == 429:
            return error.RateLimitError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 500:
            return error.APIError(error_list, rcode, rbody, resp, rheaders)
        elif rcode == 503:
            return error.ServiceUnavailableError(
                error_list, rcode, rbody, resp, rheaders
            )
        else:
            return error.APIError(error_list, rcode, rbody, resp, rheaders)

    def request_headers(self, api_key, method):
        user_agent = "Telnyx/v2 PythonBindings/%s" % (telnyx.__version__,)
        if telnyx.app_info:
            user_agent += " " + self.format_app_info(telnyx.app_info)

        ua = {
            "bindings_version": telnyx.__version__,
            "lang": "python",
            "publisher": "telnyx",
            "httplib": self._client.name,
        }
        for attr, func in [
            ["lang_version", platform.python_version],
            ["platform", platform.platform],
            ["uname", lambda: " ".join(platform.uname())],
        ]:
            try:
                val = func()
            except Exception as e:
                val = "!! %s" % (e,)
            ua[attr] = val
        if telnyx.app_info:
            ua["application"] = telnyx.app_info

        headers = {
            "X-Telnyx-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "Authorization": "Bearer %s" % (api_key,),
        }

        if (
            method == "post"
            or method == "patch"
            and headers.get("Content-Type") is None
        ):
            headers["Content-Type"] = "application/json"

        return headers

    def request_raw(self, method, url, params=None, supplied_headers=None):
        """
        Mechanism for issuing an API call
        """

        if self.api_key:
            my_api_key = self.api_key
        else:
            from telnyx import api_key

            my_api_key = api_key

        if supplied_headers is None:
            supplied_headers = {}

        if my_api_key is None:
            raise error.AuthenticationError(
                "No API key provided. (HINT: set your API key using "
                '"telnyx.api_key = <API-KEY>"). You can generate API keys '
                "from the Telnyx web interface.  See https://developers.telnyx.com/docs/v2/development/authentication "
                "for details, or email support@telnyx.com if you have any "
                "questions."
            )

        abs_url = "%s%s" % (self.api_base, url)

        if method == "get" or method == "head" or method == "delete":
            encoded_params = self.build_query_params(params)
            if params:
                abs_url = _build_api_url(abs_url, encoded_params)
            post_data = None
        elif method == "post" or method == "patch" or method == "put":
            if (
                supplied_headers.get("Content-Type", "application/json")
                == "application/json"
            ):
                encoded_params = json.dumps(params or {})
                post_data = encoded_params
            elif supplied_headers.get("Content-Type", None) == "multipart/form-data":
                generator = MultipartDataGenerator()
                generator.add_params(params or {})
                post_data = generator.get_post_data()
                encoded_params = params
                supplied_headers["Content-Type"] = (
                    "multipart/form-data; boundary=%s" % (generator.boundary,)
                )
        else:
            raise error.APIConnectionError(
                "Unrecognized HTTP method %r.  This may indicate a bug in the "
                "Telnyx bindings.  Please contact support@telnyx.com for "
                "assistance." % (method,)
            )

        headers = self.request_headers(my_api_key, method)
        if supplied_headers is not None:
            for key, value in six.iteritems(supplied_headers):
                headers[key] = value

        util.log_info("Request to Telnyx api", method=method, path=abs_url)
        util.log_debug("Post details", post_data=encoded_params)

        rbody, rcode, rheaders = self._client.request_with_retries(
            method, abs_url, headers, post_data
        )

        util.log_info("Telnyx API response", path=abs_url, response_code=rcode)
        util.log_debug("API response body", body=rbody)

        return rbody, rcode, rheaders, my_api_key

    def build_query_params(self, params):
        encoded_params = urlencode(list(_api_encode(params or {})))

        # Don't use strict form encoding by changing the square bracket control
        # characters back to their literals. This is fine by the server, and
        # makes these parameter strings easier to read.
        encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")
        return encoded_params

    def interpret_response(self, rbody, rcode, rheaders):
        if rcode == 204:
            resp = TelnyxResponse("", rcode, rheaders)
        else:
            try:
                if hasattr(rbody, "decode"):
                    rbody = rbody.decode("utf-8")
                resp = TelnyxResponse(rbody, rcode, rheaders)
            except Exception:
                raise error.APIError(
                    "Invalid response body from API: %s "
                    "(HTTP response code was %d)" % (rbody, rcode),
                    rbody,
                    rcode,
                    rheaders,
                )
            if not (200 <= rcode < 300):
                self.handle_error_response(rbody, rcode, resp.data, rheaders)

        return resp



from __future__ import absolute_import, division, print_function

from telnyx import error
from telnyx.api_resources.abstract import ListableAPIResource


class InventoryCoverage(ListableAPIResource):
    OBJECT_NAME = "inventory_coverage"

    @classmethod
    def retrieve(cls, id, api_key=None, **params):
        raise error.InvalidRequestError(
            "%s does not support retrieve()" % cls.class_url()
        )

    @classmethod
    def class_url(cls):
        return "/v2/inventory_coverage"
